from pathlib import Path

from .ccxt import CCXTProvider
from .capitalcom import CapitalComProvider

# List of available providers
available_providers = tuple(
    p.stem for p in Path(__file__).parent.resolve().glob('*.py') if
    p.name not in ('__init__.py', 'provider.py')
)



from typing import Callable
from abc import abstractmethod
from pathlib import Path
from datetime import datetime
import tomllib

from ..types.ohlcv import OHLCV
from pynecore.core.syminfo import SymInfo, SymInfoInterval, SymInfoSession
from pynecore.core.ohlcv_file import OHLCVWriter, OHLCVReader


class Provider:
    """
    Base class for all providers
    """

    timezone = 'UTC'
    """ Timezone of the provider """

    symbol: str | None = None
    """ Symbol of the provider """

    timeframe: str | None = None
    """ Timeframe of the provider """

    xchg_timeframe: str | None = None
    """ TradingView timeframe """

    ohlcv_path: Path | None = None
    """ Directory to save OHLV data """

    config_keys = {
        '# Settings for the provider': '',
    }
    """ Key-value pairs to put into providers.toml, if key starts with '#' it is a comment. """

    config: dict[str, str] = {}
    """ Config dict for the exchange loaded from providers.toml """

    @classmethod
    @abstractmethod
    def to_tradingview_timeframe(cls, timeframe: str) -> str:
        """
        Convert timeframe to TradingView fmt
        https://www.tradingview.com/pine-script-reference/v6/#var_timeframe.period
        """

    @classmethod
    @abstractmethod
    def to_exchange_timeframe(cls, timeframe: str) -> str:
        """
        Convert timeframe to exchange fmt
        """

    @classmethod
    def get_ohlcv_path(cls, symbol: str, timeframe: str, ohlv_dir: Path, provider_name: str | None = None) -> Path:
        """
        Get the output path of the OHLV data
        """
        return ohlv_dir / (f"{provider_name or cls.__name__.lower().replace('provider', '')}"
                           f"_{symbol.replace('/', '_').replace(':', '_')}"
                           f"_{timeframe}.ohlcv")

    def __init__(self, *, symbol: str | None = None, timeframe: str | None = None,
                 ohlv_dir: Path | None = None, config_dir: Path | None = None):
        """
        :param symbol: The symbol to get data for
        :param timeframe: The timeframe to get data for in TradingView fmt
        :param ohlv_dir: The directory to save OHLV data
        :param config_dir: The directory to read the config file from
        """
        self.symbol = symbol
        self.timeframe = timeframe
        self.xchg_timeframe = self.to_exchange_timeframe(timeframe) if timeframe else None
        self.ohlcv_path = self.get_ohlcv_path(symbol, timeframe, ohlv_dir) if ohlv_dir else None
        self.ohlcv_file = OHLCVWriter(self.ohlcv_path) if self.ohlcv_path else None

        if not config_dir:  # Default config dir from the parent of the ohlcv_dir
            assert self.ohlcv_path is not None
            config_dir = self.ohlcv_path.parent.parent / 'config'
        self.config_dir = config_dir

        self.load_config()

    def __enter__(self) -> OHLCVWriter:
        assert self.ohlcv_file is not None
        return self.ohlcv_file.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert self.ohlcv_file is not None
        self.ohlcv_file.close()

    @abstractmethod
    def get_list_of_symbols(self, *args, **kwargs) -> list[str]:
        """
        Get list of symbols
        """

    def load_config(self):
        """
        Load config from providers.toml
        """
        with open(self.config_dir / 'providers.toml', 'rb') as f:
            data = tomllib.load(f)
            self.config = data[self.__class__.__name__.replace('Provider', '').lower()]

    @abstractmethod
    def update_symbol_info(self) -> SymInfo:
        """
        Update symbol info from the exchange
        """

    def is_symbol_info_exists(self) -> bool:
        """
        Check if symbol info file exists
        """
        assert self.ohlcv_path is not None
        return self.ohlcv_path.with_suffix('.toml').exists()

    def get_symbol_info(self, force_update=False) -> SymInfo:
        """
        Get market details of a symbol

        :param force_update: Force update the symbol info
        """
        assert self.ohlcv_path is not None
        toml_path = self.ohlcv_path.with_suffix('.toml')
        # Check if file already exists
        if self.is_symbol_info_exists() and not force_update:
            return SymInfo.load_toml(toml_path)

        sym_info = self.update_symbol_info()
        sym_info.save_toml(toml_path)
        return sym_info

    @abstractmethod
    def get_opening_hours_and_sessions(self) \
            -> tuple[list[SymInfoInterval], list[SymInfoSession], list[SymInfoSession]]:
        """
        Get opening hours and sessions of a symbol
        """

    def save_ohlcv_data(self, data: OHLCV | list[OHLCV]):
        """
        Save OHLV data to a file

        :param data: OHLV data
        """
        assert self.ohlcv_file is not None
        if isinstance(data, OHLCV):
            self.ohlcv_file.write(data)
        else:
            for candle in data:
                self.ohlcv_file.write(candle)

    @abstractmethod
    def download_ohlcv(self, time_from: datetime, time_to: datetime,
                       on_progress: Callable[[datetime], None] | None = None):
        """
        Download OHLV data

        In the user code you can call `self.save_ohlcv_data()` to save the data into the data file

        :param time_from: The start time
        :param time_to: The end time
        :param on_progress: Optional callback to call on progress
        """

    def load_ohlcv_data(self) -> OHLCVReader:
        """
        Load OHLV data from the file
        """
        return OHLCVReader(str(self.ohlcv_path))



from datetime import datetime, timedelta

from rich.text import Text
from rich.progress import ProgressColumn


class DateColumn(ProgressColumn):
    """ Custom progress column showing current date being processed. """

    def __init__(self, start_time: datetime):
        super().__init__()
        self.start_time = start_time

    def render(self, task) -> Text:
        # Calculate current date from progress
        current_date = self.start_time + timedelta(seconds=task.completed)
        return Text(current_date.strftime("%Y-%m-%d %H:%M"), style="magenta")



from .series import Series
from .na import NA


class Source(Series[float]):
    """
    Represents a built-in source like "open", "high", "low", "close", "hl2", etc.
    """

    # noinspection PyMissingConstructor
    def __init__(self, name: str) -> None:
        self.name = name

    def __getitem__(self, index: int) -> float | NA[float]:
        return NA(float)



from pathlib import Path
from datetime import datetime

from typer import Option, Argument, secho, Exit
from rich.progress import (Progress, SpinnerColumn, TextColumn, BarColumn,
                           TimeElapsedColumn, TimeRemainingColumn)

from ..app import app, app_state

from ...utils.rich.date_column import DateColumn
from pynecore.core.ohlcv_file import OHLCVReader

from pynecore.core.syminfo import SymInfo
from pynecore.core.script_runner import ScriptRunner

__all__ = []


@app.command()
def run(
        script: Path = Argument(..., dir_okay=False, file_okay=True, help="Script to run"),
        data: Path = Argument(..., dir_okay=False, file_okay=True,
                              help="Data file to use (*.ohlcv)"),
        time_from: datetime | None = Option(None, '--from', '-f',
                                            formats=["%Y-%m-%d", "%Y-%m-%d %H:%M:%S"],
                                            help="Start date (UTC), if not specified, will use the "
                                                 "first date in the data"),
        time_to: datetime | None = Option(None, '--to', '-t',
                                          formats=["%Y-%m-%d", "%Y-%m-%d %H:%M:%S"],
                                          help="End date (UTC), if not specified, will use the last "
                                               "date in the data"),
        plot_path: Path | None = Option(None, "--plot", "-pp",
                                        help="Path to save the plot data",
                                        rich_help_panel="Out Path Options"),
        strat_path: Path | None = Option(None, "--strat", "-sp",
                                         help="Path to save the strategy statistics",
                                         rich_help_panel="Out Path Options"
                                         ),
        equity_path: Path | None = Option(None, "--equity", "-ep",
                                          help="Path to save the equity curve",
                                          rich_help_panel="Out Path Options"),
):
    """
    Run a script

    The system automatically searches for the workdir folder in the current and parent directories.
    If not found, it creates or uses a workdir folder in the current directory.

    If [bold]script[/] path is a name without full path, it will be searched in the [italic]"workdir/scripts"[/] directory.
    Similarly, if [bold]data[/] path is a name without full path, it will be searched in the [italic]"workdir/data"[/] directory.
    The [bold]plot_path[/], [bold]strat_path[/], and [bold]equity_path[/] work the same way - if they are names without full paths,
    they will be saved in the [italic]"workdir/output"[/] directory.
    """  # noqa
    # Ensure .py extension
    if script.suffix != ".py":
        script = script.with_suffix(".py")
    # Expand script path
    if len(script.parts) == 1:
        script = app_state.scripts_dir / script
    # Check if script exists
    if not script.exists():
        secho(f"Script file '{script}' not found!", fg="red", err=True)
        raise Exit(1)

    # Ensure .ohlcv extension
    if data.suffix != ".ohlcv":
        data = data.with_suffix(".ohlcv")
    # Expand data path
    if len(data.parts) == 1:
        data = app_state.data_dir / data
    # Check if data exists
    if not data.exists():
        secho(f"Data file '{data}' not found!", fg="red", err=True)
        raise Exit(1)

    # Ensure .csv extension for plot path
    if plot_path and plot_path.suffix != ".csv":
        plot_path = plot_path.with_suffix(".csv")
    if not plot_path:
        plot_path = app_state.output_dir / f"{script.stem}.csv"

    # Ensure .csv extension for strategy path
    if strat_path and strat_path.suffix != ".csv":
        strat_path = strat_path.with_suffix(".csv")
    if not strat_path:
        strat_path = app_state.output_dir / f"{script.stem}_strat.csv"

    # Ensure .csv extension for equity path
    if equity_path and equity_path.suffix != ".csv":
        equity_path = equity_path.with_suffix(".csv")
    if not equity_path:
        equity_path = app_state.output_dir / f"{script.stem}_equity.csv"

    # Get symbol info for the data
    try:
        syminfo = SymInfo.load_toml(data.with_suffix(".toml"))
    except FileNotFoundError:
        secho(f"Symbol info file '{data.with_suffix('.toml')}' not found!", fg="red", err=True)
        raise Exit(1)

    # Open data file
    with OHLCVReader(data) as reader:
        if not time_from:
            time_from = reader.start_datetime
        if not time_to:
            time_to = reader.end_datetime
        time_from = time_from.replace(tzinfo=None)
        time_to = time_to.replace(tzinfo=None)

        total_seconds = int((time_to - time_from).total_seconds())

        # Get the iterator
        size = reader.get_size(int(time_from.timestamp()), int(time_to.timestamp()))
        ohlcv_iter = reader.read_from(int(time_from.timestamp()), int(time_to.timestamp()))

        with Progress(
                SpinnerColumn(finished_text="[green]✓"),
                TextColumn("{task.description}"),
                DateColumn(time_from),
                BarColumn(),
                TimeElapsedColumn(),
                "/",
                TimeRemainingColumn(),
        ) as progress:
            task = progress.add_task(
                description="Running script...",
                total=total_seconds,
            )

            def cb_progress(current_time: datetime | None):
                """ Callback to update progress """
                if current_time == datetime.max:
                    current_time = time_to
                assert current_time is not None
                elapsed_seconds = int((current_time - time_from).total_seconds())
                progress.update(task, completed=elapsed_seconds)

            # Create script runner
            runner = ScriptRunner(script, ohlcv_iter, syminfo, last_bar_index=size - 1,
                                  plot_path=plot_path, strat_path=strat_path, equity_path=equity_path)
            # Run the script
            runner.run(on_progress=cb_progress)



from ..types.display import Display

all: Display = Display()  # noqa F821
data_window: Display = Display()
none: Display = Display()
panel: Display = Display()
price_scale: Display = Display()
status_line: Display = Display()



from ..types.plotstyle import PlotStyle

#
# Constants
#

style_area = PlotStyle()
style_areabr = PlotStyle()
style_circles = PlotStyle()
style_columns = PlotStyle()
style_cross = PlotStyle()
style_histogram = PlotStyle()
style_line = PlotStyle()
style_linebr = PlotStyle()
style_stepline = PlotStyle()
style_stepline_diamond = PlotStyle()



import ast


class PersistentSeriesTransformer(ast.NodeTransformer):
    """
    Transform PersistentSeries declarations into Persistent + Series combination.
    Must be applied before PersistentTransformer and SeriesTransformer.
    """

    def visit_ImportFrom(self, node):
        """Handle imports, only remove Series while keeping other imports"""
        if node.module and node.module.startswith('pynecore'):
            # Filter out Persistent from names
            new_names = [name for name in node.names if name.name != 'PersistentSeries']
            if not new_names:
                # If no names left, remove the entire import
                return None
            # Create new import with remaining names
            node.names = new_names
        return node

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.AST | list[ast.AnnAssign]:
        """Transform PersistentSeries type annotations into separate Persistent and Series declarations"""
        if hasattr(node, '_ps_transformed'):
            return node

        if not isinstance(node.target, ast.Name):
            return node

        # Check if it's a PersistentSeries type
        is_persistent_series = False
        series_type = None

        if isinstance(node.annotation, ast.Subscript):
            if (isinstance(node.annotation.value, ast.Name) and
                    node.annotation.value.id == 'PersistentSeries'):
                is_persistent_series = True
                series_type = node.annotation.slice
        elif (isinstance(node.annotation, ast.Name) and
              node.annotation.id == 'PersistentSeries'):
            is_persistent_series = True

        if not is_persistent_series:
            return node

        # Create two declarations
        var_name = node.target.id
        value = node.value

        # 1. Persistent declaration
        persistent_decl = ast.AnnAssign(
            target=ast.Name(id=var_name, ctx=ast.Store()),
            annotation=ast.Subscript(
                value=ast.Name(id='Persistent', ctx=ast.Load()),
                slice=series_type if series_type else ast.Name(id='float', ctx=ast.Load()),
                ctx=ast.Load()
            ) if series_type else ast.Name(id='Persistent', ctx=ast.Load()),
            value=value,
            simple=1
        )
        persistent_decl._ps_transformed = True

        # 2. Series declaration
        series_decl = ast.AnnAssign(
            target=ast.Name(id=var_name, ctx=ast.Store()),
            annotation=ast.Subscript(
                value=ast.Name(id='Series', ctx=ast.Load()),
                slice=series_type if series_type else ast.Name(id='float', ctx=ast.Load()),
                ctx=ast.Load()
            ) if series_type else ast.Name(id='Series', ctx=ast.Load()),
            value=ast.Name(id=var_name, ctx=ast.Load()),
            simple=1
        )
        series_decl._ps_transformed = True

        return [persistent_decl, series_decl]



from typing import Callable
import inspect
import sys

__all__ = ['export']

_do_nothing = lambda: None


def export(*to_export: Callable | str) -> Callable:
    """
    Decorator to add function name into its module's __all__ list
    It can also be used as a function to export a list of names or a single name

    :param to_export: The function to export or a list of names to export
    :return: The decorated function (the same function)
    """
    assert to_export, "@export decorator must be called with a function or a list of names to export!"
    to_export = list(to_export)  # type: ignore
    assert callable(to_export[0]) or all(isinstance(name, str) for name in to_export), \
        "@export decorator can only be used on function or list of strings of module attributes!"

    # Get the from the decorator called
    caller = inspect.stack()[1]
    module = sys.modules[caller.frame.f_globals['__name__']]
    if not hasattr(module, '__all__'):
        module.__all__ = []  # type: ignore

    # Function
    if len(to_export) == 1 and callable(to_export[0]):
        module.__all__.append(to_export[0].__name__)
        return to_export[0]

    # List of names
    for name in to_export:
        module.__all__.append(name)

    # Return do_nothing function
    return _do_nothing



from datetime import datetime, timedelta

from ..types.session import Session

from ..core.module_property import module_property
from ..utils.export import export

from . import syminfo
from . import timeframe
from .. import lib

#
# Constants
#

regular = Session()
extended = Session()

export("regular", "extended")


#
# Functions
#

# noinspection PyProtectedMember
def _check_session(dt: datetime, tf_sec: int) -> bool:
    """
    Check if candle overlaps with any trading session.

    :param dt: Start datetime of the candle
    :param tf_sec: Timeframe in seconds
    :return: True if candle overlaps with any session
    """
    candle_end = dt + timedelta(seconds=tf_sec)
    candle_start = dt

    for day, ss, se in syminfo._opening_hours:
        if day != dt.weekday():
            continue

        ssdt = dt.replace(hour=ss.hour, minute=ss.minute, second=ss.second, microsecond=0)
        sedt = dt.replace(hour=se.hour, minute=se.minute, second=se.second, microsecond=0)
        if sedt < ssdt:  # Overnight session
            sedt += timedelta(days=1)

        # Check if candle overlaps with session
        if candle_end >= ssdt and candle_start < sedt:
            return True

    return False


#
# Module properties
#

# noinspection PyProtectedMember
@export
@module_property
def isfirstbar_regular() -> bool:
    """
    Check if the current candle is the first of the trading session.
    The result is the same whether extended session information is used or not.

    :return: True if the current candle is the first of the trading session
    """
    tf_sec = timeframe.in_seconds(syminfo.period)
    for ss in syminfo._session_starts:
        if ss.day == lib._datetime.weekday():
            ssdt = lib._datetime.replace(hour=ss.time.hour, minute=ss.time.minute, second=ss.time.second,
                                         microsecond=ss.time.microsecond)
            if lib._datetime <= ssdt < lib._datetime + timedelta(seconds=tf_sec):
                return True
    return False


# noinspection PyProtectedMember
# TODO: implement this when extended session will be supported
@export
@module_property
def isfirstbar() -> bool:
    """
    Check if the current candle is the first of the trading session.
    If extended session information is used, only returns true on the first bar of the pre-market bars.
    NOTE: extended session information is not yet supported.

    :return: True if the current candle is the first of the trading session
    """
    # TODO: support pre market sessions
    tf_sec = timeframe.in_seconds(syminfo.period)
    for ss in syminfo._session_starts:
        if ss.day == lib._datetime.weekday():
            ssdt = lib._datetime.replace(hour=ss.time.hour, minute=ss.time.minute, second=ss.time.second,
                                         microsecond=ss.time.microsecond)
            if lib._datetime <= ssdt < lib._datetime + timedelta(seconds=tf_sec):
                return True
    return False


# noinspection PyProtectedMember
@export
@module_property
def islastbar_regular() -> bool:
    """
    Check if the current candle is the last of the trading session.
    The result is the same whether extended session information is used or not.

    :return: True if the current candle is the last of the trading session
    """
    tf_sec = timeframe.in_seconds(syminfo.period)
    for se in syminfo._session_ends:
        if se.day == lib._datetime.weekday():
            sedt = lib._datetime.replace(hour=se.time.hour, minute=se.time.minute, second=se.time.second,
                                         microsecond=se.time.microsecond)
            if lib._datetime < sedt < lib._datetime + timedelta(seconds=tf_sec):
                return True
    return False


# noinspection PyProtectedMember
@export
@module_property
def islastbar() -> bool:
    """
    Check if the current candle is the last of the trading session.
    If extended session information is used, only returns true on the last bar of the post-market bars.
    NOTE: extended session information is not yet supported.

    :return: True if the current candle is the last of the trading session
    """
    tf_sec = timeframe.in_seconds(syminfo.period)
    for se in syminfo._session_ends:
        if se.day == lib._datetime.weekday():
            sedt = lib._datetime.replace(hour=se.time.hour, minute=se.time.minute, second=se.time.second,
                                         microsecond=se.time.microsecond)
            if lib._datetime < sedt < lib._datetime + timedelta(seconds=tf_sec):
                return True
    return False


# noinspection PyProtectedMember
@export
@module_property
def ismarket() -> bool:
    """
    Check if the current candle is within a trading session.

    :return:  True if the current candle is within a trading session
    """
    tf_sec = timeframe.in_seconds(syminfo.period)
    return _check_session(lib._datetime, tf_sec)


@export
@module_property
def ispremarket() -> bool:
    """
    Check if the current candle is within the pre-market session.
    It is not yet implemented.

    :return: It is always False at the moment
    """
    # TODO: implement this
    return False


@export
@module_property
def ispostmarket() -> bool:
    """
    Check if the current candle is within the post-market session.
    It is not yet implemented.

    :return: It is always False at the moment
    """
    # TODO: implement this
    return False



from ..types.shape import Shape

#
# Constants
#

arrowdown = Shape()
arrowup = Shape()
circle = Shape()
cross = Shape()
diamond = Shape()
flag = Shape()
labeldown = Shape()
labelup = Shape()
square = Shape()
triangledown = Shape()
triangleup = Shape()
xcross = Shape()



from __future__ import annotations
from typing import TypeVar, overload, Protocol, Any, Generic

T = TypeVar('T')


class ModuleProperyProtocol(Protocol[T]):
    @overload
    def __call__(self) -> T: ...

    @overload
    def __call__(self, *args: Any, **kwargs: Any) -> T: ...

    def __call__(self, *args: Any, **kwargs: Any) -> T: ...


def module_property(func) -> ModuleProperyProtocol[T] | T:
    """
    Decorator for Pine-style hybrid property/functions.
    """
    setattr(func, '__module_property__', True)
    return func



from functools import lru_cache
from datetime import datetime, timedelta, UTC

from ..core.module_property import module_property
from ..utils.export import export

from .. import lib
from . import syminfo as _syminfo
from pynecore.core.datetime import parse_timezone as _parse_timezone

__persistent_function_vars__ = {}


@lru_cache(maxsize=128)
def _process_tf(timeframe: str) -> tuple[str, int]:
    """
    Process the timeframe string and return the modi    fier and multiplier

    :param timeframe: The timeframe string in TradingView format
    :return: A tuple with the modifier and multiplier
    :raises AssertionError: If the timeframe is invalid
    """
    assert len(timeframe) > 0, "Invalid timeframe: empty string!"

    # Simple minutes
    if timeframe.isdigit():
        _modifier = ''
        _multiplier = int(timeframe)

    # Multiplier and modifier
    elif len(timeframe) > 1:
        if not timeframe[-1].isdigit():
            _modifier = timeframe[-1]
            _multiplier = int(timeframe[:-1])
        else:
            raise AssertionError("Invalid timeframe format!")

    # Just a single character
    else:
        _modifier = timeframe
        _multiplier = 1

    assert _modifier in ('', 'T', 'S', 'D', 'W', 'M'), "Invalid timeframe: wrong modifier!"
    assert _multiplier > 0, "Invalid timeframe: wrong multiplier!"
    return _modifier, _multiplier


def _is_time_in_candle(candle_time: datetime, check_time: datetime, tf_sec: int) -> bool:
    """
    Check if a given time is within the candle's timeframe.

    :param candle_time: Start datetime of the candle
    :param check_time: Time to check
    :param tf_sec: Timeframe in seconds
    :return: True if the time is within the candle
    """
    candle_end = candle_time + timedelta(seconds=tf_sec)
    return candle_time <= check_time < candle_end


# noinspection PyProtectedMember
def _get_first_session_start_in_day(dt: datetime) -> datetime | None:
    """
    Get first session start time in the given day using session_starts configuration.

    :param dt: The datetime to check (in UTC)
    :return: First session start time in UTC
    """
    local_dt = dt.astimezone(_parse_timezone(_syminfo.timezone))
    weekday = local_dt.weekday()

    # Get all session starts for current weekday
    day_starts = [
        time for day, time in _syminfo._session_starts
        if day == weekday
    ]

    if not day_starts:
        return None

    # Get the earliest session start for the day
    first_start = min(day_starts)

    # Create datetime with the session start time
    ssdt = local_dt.replace(
        hour=first_start.hour,
        minute=first_start.minute,
        second=first_start.second,
        microsecond=0
    )

    return ssdt.astimezone(UTC)


# noinspection PyProtectedMember
def _get_new_year_session(dt: datetime) -> datetime | None:
    """
    Get the new year session of the given datetime.

    :param dt: The datetime to start searching in UTC
    :return: The next new year session in exchange timezone
    """
    dt_utc = dt.astimezone(UTC)
    nydt_utc = dt_utc.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
    # Get next new year
    while True:
        ss = _get_first_session_start_in_day(nydt_utc)
        if ss is None:
            nydt_utc += timedelta(days=1)
            continue
        # Found the next new year session, which will be the next anchor
        return ss.astimezone(_parse_timezone(_syminfo.timezone))


# noinspection PyProtectedMember
def _is_new_session(current_dt: datetime, prev_dt: datetime | None = None, tf_sec: int | None = None) -> bool:
    """
    Check if current bar starts a new session.

    :param current_dt: Current candle datetime (in local exchange timezone)
    :param prev_dt: Previous candle datetime (in local exchange timezone)
    :param tf_sec: Timeframe width in seconds
    :return: True if this is the first candle of a new session
    """
    if tf_sec is None:
        tf_sec = in_seconds(_syminfo.period)
    if prev_dt is None:
        prev_dt = current_dt - timedelta(seconds=tf_sec)

    current_weekday = current_dt.weekday()
    prev_weekday = prev_dt.weekday()

    # Get all possible session starts:
    # 1. Current day sessions
    # 2. If previous day different, then previous day's overnight sessions
    session_starts = [
        ss for day, ss in _syminfo._session_starts
        if day == current_weekday
    ]

    if prev_weekday != current_weekday:
        prev_day_starts = [
            (ss, se) for day, ss, se in _syminfo._opening_hours
            if day == prev_weekday and se < ss  # Only overnight sessions
        ]
        # Add overnight session starts to check
        for ss, se in prev_day_starts:
            session_starts.append(ss)

    # For each session start
    for start_time in session_starts:
        session_start = current_dt.replace(
            hour=start_time.hour,
            minute=start_time.minute,
            second=start_time.second,
            microsecond=0
        )

        # If it crosses the day boundary, we set it to the previous day
        if prev_weekday != current_weekday and start_time > current_dt.time():
            session_start = session_start - timedelta(days=1)

        # If session starts in this bar
        if (current_dt <= session_start < current_dt + timedelta(seconds=tf_sec) and
                session_start > prev_dt):
            return True

    return False


__persistent_next_new_year_session: datetime | None = None
__persistent_cycle: int = 0
__persistent_last_dt: datetime | None = None
__persistent_last_signal: datetime | None = None
__persistent_next_anchor: datetime | None = None
__persistent_next_signal: datetime | None = None
__persistent_function_vars__['change'] = ['__persistent_next_new_year_session', '__persistent_cycle',
                                          '__persistent_last_dt', '__persistent_next_anchor',
                                          '__persistent_next_signal', '__persistent_last_signal']


# noinspection PyProtectedMember
# TODO: make it simpler and better if you could
# I know this function is awful. It was one of the hardest part of the whole Pine library.
# It may be simplified. The problem is that every timefram has different anchor points and slightly different rules. Or
# just not found the general rule for all timeframes.
@export
def change(timeframe: str) -> bool:
    """
    Detects changes in the specified timeframe.

    :param timeframe: The timeframe to check
    :return: Returns true on the first bar of a `timeframe`, false otherwise.
    """
    global __persistent_next_new_year_session, __persistent_cycle, __persistent_last_dt, \
        __persistent_next_anchor, __persistent_next_signal, __persistent_last_signal

    tf_sec = in_seconds(timeframe)
    xchg_tf = _syminfo.period
    xchg_tf_sec = in_seconds(xchg_tf)

    # The timeframe to check nust be greater (or equal) than the current timeframe
    if tf_sec < xchg_tf_sec:
        return False

    _modifier, _multiplier = _process_tf(timeframe)
    assert _modifier != 'T', "Ticks are not (yet) supported!"
    is_intraday = _modifier == '' or _modifier == 'S'

    # Datetime in exchange timezone
    dt: datetime = lib._datetime

    # Initialize persistent variables
    if __persistent_last_dt is None:
        __persistent_last_dt = dt - timedelta(seconds=xchg_tf_sec)
        __persistent_last_signal = __persistent_last_dt

    # Find anchor point and replay sessions to the 1st bar
    if not is_intraday and __persistent_next_new_year_session is None:
        dt_utc = dt.astimezone(UTC)
        nydt_utc = dt_utc.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        nydt = _get_new_year_session(nydt_utc)

        assert nydt is not None

        __persistent_next_new_year_session = _get_new_year_session(dt_utc.replace(year=dt_utc.year + 1))
        __persistent_cycle = 0

        # Daily timeframe's anchor is the first session of the year
        if _modifier == 'D':
            _dt = nydt

            while _dt < dt:
                if _is_new_session(_dt, _dt - timedelta(seconds=xchg_tf_sec), xchg_tf_sec):
                    __persistent_cycle -= 1
                    if __persistent_cycle <= 0:
                        __persistent_last_signal = _dt
                        __persistent_cycle = _multiplier

                _dt += timedelta(seconds=xchg_tf_sec)

        # Weekly timeframe's anchor is the first Monday of the year
        elif _modifier == 'W':
            _dt_utc = nydt_utc

            # The anchor point must be Monday
            while _dt_utc < dt_utc:
                if _dt_utc.weekday() != 0:
                    _dt_utc += timedelta(days=1)
                    continue
                break

            # Find the next signal
            while _dt_utc < dt_utc:
                _dt_utc += timedelta(seconds=tf_sec)

            __persistent_next_anchor = _dt_utc
            __persistent_next_signal = None
            while __persistent_next_signal is None:
                __persistent_next_signal = _get_first_session_start_in_day(_dt_utc)
                _dt_utc += timedelta(days=1)
            __persistent_next_signal = __persistent_next_signal.astimezone(_parse_timezone(_syminfo.timezone))

        # Monthly timeframe's anchor is the first day of the month
        elif _modifier == 'M':
            if dt_utc.month == 12:
                next_month_dt = dt_utc.replace(year=dt_utc.year + 1, month=1, day=1, hour=0, minute=0, second=0)
            else:
                next_month_dt = dt_utc.replace(month=dt_utc.month + 1, day=1, hour=0, minute=0, second=0)

            __persistent_next_signal = None
            while __persistent_next_signal is None:
                __persistent_next_signal = _get_first_session_start_in_day(next_month_dt)
                next_month_dt += timedelta(days=1)
            __persistent_next_signal = __persistent_next_signal.astimezone(_parse_timezone(_syminfo.timezone))

            __persistent_cycle = dt_utc.month

    # We need to check every virtual candles, even if they are missing in the dataset
    while __persistent_last_dt < dt:
        prev_dt = __persistent_last_dt
        __persistent_last_dt += timedelta(seconds=xchg_tf_sec)

        if is_intraday:
            # The anchor point is the session start
            if _is_new_session(__persistent_last_dt, prev_dt, xchg_tf_sec):
                # We need to round the session start to the nearest hour
                __persistent_last_signal = __persistent_last_dt.replace(minute=0, second=0, microsecond=0)
            assert __persistent_last_signal is not None
            seconds_since_last_session = (__persistent_last_dt - __persistent_last_signal).total_seconds()

            if seconds_since_last_session % tf_sec == 0:
                __persistent_last_dt = dt
                return lib.bar_index > 0

        # We need to check only trading days
        elif _modifier == 'D':
            # Check if it is a new year
            assert __persistent_next_new_year_session is not None
            if (_is_time_in_candle(__persistent_last_dt, __persistent_next_new_year_session, xchg_tf_sec) or
                    __persistent_next_new_year_session < __persistent_last_dt):
                # Find the next new year session
                __persistent_next_new_year_session = _get_new_year_session(
                    __persistent_next_new_year_session + timedelta(days=367))
                __persistent_cycle = 0

            if _is_new_session(__persistent_last_dt, prev_dt, xchg_tf_sec):
                __persistent_cycle -= 1
                assert __persistent_last_signal is not None
                if __persistent_cycle <= 0:
                    __persistent_cycle = _multiplier
                    if __persistent_last_signal < __persistent_last_dt:
                        __persistent_last_signal = dt
                        return lib.bar_index > 0

        # We don't need to skip weekends here
        elif _modifier == 'W':
            assert __persistent_next_new_year_session is not None
            if (_is_time_in_candle(__persistent_last_dt, __persistent_next_new_year_session, xchg_tf_sec) or
                    __persistent_next_new_year_session < __persistent_last_dt):
                _dt_utc = ((__persistent_next_new_year_session + timedelta(days=1))
                           .astimezone(UTC).replace(day=1, month=1, hour=0, minute=0, second=0))
                __persistent_next_new_year_session = _get_new_year_session(_dt_utc.replace(year=_dt_utc.year + 1))

                # The anchor point must be Monday
                while True:
                    if _dt_utc.weekday() != 0:
                        _dt_utc += timedelta(days=1)
                        continue
                    break

                # 1st signal of the new year
                __persistent_next_signal = None
                while __persistent_next_signal is None:
                    __persistent_next_signal = _get_first_session_start_in_day(_dt_utc)
                    _dt_utc += timedelta(days=1)
                __persistent_next_anchor = __persistent_next_signal
                __persistent_next_signal = __persistent_next_signal.astimezone(_parse_timezone(_syminfo.timezone))

            assert __persistent_next_signal is not None
            if (_is_time_in_candle(__persistent_last_dt, __persistent_next_signal, xchg_tf_sec)
                    or __persistent_last_dt >= __persistent_next_signal):
                assert __persistent_next_anchor is not None
                # Increase the anchor date
                __persistent_next_anchor += timedelta(seconds=tf_sec)
                # Find the next signal which is usually Monday
                _dt_utc = __persistent_next_anchor
                __persistent_next_signal = None
                while __persistent_next_signal is None:
                    __persistent_next_signal = _get_first_session_start_in_day(_dt_utc)
                    _dt_utc += timedelta(days=1)
                __persistent_next_signal = __persistent_next_signal.astimezone(_parse_timezone(_syminfo.timezone))
                return lib.bar_index > 0

        elif _modifier == 'M':
            # Is it a new year candle?
            assert __persistent_next_new_year_session is not None
            if (_is_time_in_candle(__persistent_last_dt, __persistent_next_new_year_session, xchg_tf_sec) or
                    __persistent_next_new_year_session < __persistent_last_dt):
                _dt_utc = ((__persistent_next_new_year_session + timedelta(days=1))
                           .astimezone(UTC).replace(day=1, month=1, hour=0, minute=0, second=0))
                __persistent_next_new_year_session = _get_new_year_session(_dt_utc.replace(year=_dt_utc.year + 1))

                # 1st signal of the new year
                __persistent_next_signal = None
                while __persistent_next_signal is None:
                    __persistent_next_signal = _get_first_session_start_in_day(_dt_utc)
                    _dt_utc += timedelta(days=1)
                __persistent_next_signal = __persistent_next_signal.astimezone(_parse_timezone(_syminfo.timezone))

                __persistent_cycle = 0

            assert __persistent_next_signal is not None
            if (_is_time_in_candle(__persistent_last_dt, __persistent_next_signal, xchg_tf_sec)
                    or __persistent_last_dt >= __persistent_next_signal):
                _dt_utc = __persistent_next_signal.astimezone(UTC) + timedelta(days=1)
                if _dt_utc.month == 12:
                    next_month_dt = _dt_utc.replace(year=_dt_utc.year + 1, month=1, day=1, hour=0, minute=0, second=0)
                else:
                    next_month_dt = _dt_utc.replace(month=_dt_utc.month + 1, day=1, hour=0, minute=0, second=0)
                __persistent_next_signal = None
                while __persistent_next_signal is None:
                    __persistent_next_signal = _get_first_session_start_in_day(next_month_dt)
                    next_month_dt += timedelta(days=1)

                __persistent_next_signal = __persistent_next_signal.astimezone(_parse_timezone(_syminfo.timezone))

                m = __persistent_cycle
                __persistent_cycle += 1

                if m % _multiplier == 0:
                    return lib.bar_index > 0

    return False


@export
@lru_cache(maxsize=128)
def from_seconds(seconds: int) -> str:
    """
    Convert seconds to a timeframe

    :param seconds: The seconds to convert
    :return: The timeframe in TradingView format
    """
    if seconds % (60 * 60 * 24 * 7 * 4) == 0:
        return f"{seconds // (60 * 60 * 24 * 7 * 4)}M"
    if seconds % (60 * 60 * 24 * 7) == 0:
        return f"{seconds // (60 * 60 * 24 * 7)}W"
    if seconds % (60 * 60 * 24) == 0:
        return f"{seconds // (60 * 60 * 24)}D"
    if seconds % (60 * 60) == 0:
        return f"{seconds // (60 * 60)}"
    return f"{seconds}S"


@export
def in_seconds(timeframe: str) -> int:
    """
    Convert the timeframe to seconds

    :param timeframe: The timeframe to convert
    :return: The timeframe in seconds
    :raises ValueError: If the timeframe is invalid
    """
    _modifier, _multiplier = _process_tf(timeframe)
    if _modifier == 'S':
        return _multiplier
    elif _modifier == 'D':
        return _multiplier * 60 * 60 * 24
    elif _modifier == 'W':
        return _multiplier * 60 * 60 * 24 * 7
    elif _modifier == 'M':
        return int(_multiplier * (60 * 60 * 24 * (365 / 12) + 3))  # Don't know why, but TV adds 3 secs here
    elif _modifier == '':
        return _multiplier * 60
    else:
        raise ValueError("Not supported timeframe!")


@export
@module_property
def isdaily() -> bool:
    """
    Check if the current timeframe is daily

    :return: True if the current timeframe is daily
    """
    modifier, _ = _process_tf(_syminfo.period)
    return modifier == 'D'


@export
@module_property
def isdwm() -> bool:
    """
    Check if the current timeframe is intraday, daily, weekly or monthly

    :return: True if the current timeframe is intraday, daily, weekly or monthly
    """
    modifier, _ = _process_tf(_syminfo.period)
    return modifier == 'D' or modifier == 'W' or modifier == 'M'


@export
@module_property
def isintraday() -> bool:
    """
    Check if the current timeframe is intraday

    :return: True if the current timeframe is intraday
    """
    modifier, _ = _process_tf(_syminfo.period)
    return modifier == '' or modifier == 'S' or modifier == 'T'


@export
@module_property
def isminutes() -> bool:
    """
    Check if the current timeframe is minutes

    :return: True if the current timeframe is minutes
    """
    modifier, _ = _process_tf(_syminfo.period)
    return modifier == ''


@export
@module_property
def ismonthly() -> bool:
    """
    Check if the current timeframe is monthly

    :return: True if the current timeframe is monthly
    """
    modifier, _ = _process_tf(_syminfo.period)
    return modifier == 'M'


@export
@module_property
def isseconds() -> bool:
    """
    Check if the current timeframe is seconds

    :return: True if the current timeframe is seconds
    """
    modifier, _ = _process_tf(_syminfo.period)
    return modifier == 'S'


@export
@module_property
def isticks() -> bool:
    """
    Check if the current timeframe is ticks

    :return: True if the current timeframe is ticks
    """
    modifier, _ = _process_tf(_syminfo.period)
    return modifier == 'T'


@export
@module_property
def isweekly() -> bool:
    """
    Check if the current timeframe is weekly

    :return: True if the current timeframe is weekly
    """
    modifier, _ = _process_tf(_syminfo.period)
    return modifier == 'W'


# noinspection PyProtectedMember
@export
@module_property
def main_period() -> str:
    """
    Get the main period

    :return: The main period
    """
    assert lib._script is not None
    return lib._script.timeframe or str(_syminfo.period)


@export
@module_property
def multiplier() -> int:
    """
    Get the current timeframe multiplier

    :return: The current timeframe multiplier
    """
    _, _multiplier = _process_tf(_syminfo.period)
    return _multiplier


@export
@module_property
def period() -> str:
    """
    Get the current period

    :return: The current period
    """
    return str(_syminfo.period)



from ..types.color import Color

#
# Constants
#

aqua = Color('#00BCD4')
black = Color('#363A45')
blue = Color('#2962ff')
fuchsia = Color('#E040FB')
gray = Color('#787B86')
green = Color('#4CAF50')
lime = Color('#00E676')
maroon = Color('#880E4F')
navy = Color('#311B92')
olive = Color('#808000')
orange = Color('#FF9800')
purple = Color('#9C27B0')
red = Color('#F23645')
silver = Color('#B2B5BE')
teal = Color('#089981')
white = Color('#FFFFFF')
yellow = Color('#FDD835')


def r(color: Color) -> int:
    """
    Return the red component of a color

    :param color: Color
    :return: The red component of the color
    """
    return color.r


def g(color: Color) -> int:
    """
    Return the green component of a color

    :param color: Color
    :return: The green component of the color
    """
    return color.g


def b(color: Color) -> int:
    """
    Return the blue component of a color

    :param color: Color
    :return: The blue component of the color
    """
    return color.b


def t(color: Color) -> int:
    """
    Return the transparency of a color

    :param color: Color
    :return: The transparency of the color, 0-100 (0: not transparent, 100: invisible)
    """
    return color.a


# noinspection PyShadowingNames
def new(color: Color | str, transp: float = 0) -> Color:
    """
    Return a new color with the same RGB values and a different transparency

    :param color: A color object or a string in "#RRGGBB" or "#RRGGBBAA" format
    :param transp: Transparency percentage (0-100, 0: not transparent, 100: invisible)
    """
    if isinstance(color, str):
        color = Color(color)
    color.t = transp
    return color


# noinspection PyShadowingNames
def rgb(r: int, g: int, b: int, transp: float = 0) -> Color:
    """
    Return a new color with the given RGB values and transparency

    :param r: Red value
    :param g: Green value
    :param b: Blue value
    :param transp: Transparency percentage (0-100, 0: not transparent, 100: invisible)
    """
    return Color.rgb(r, g, b, transp)



from typing import cast
import os
import sys
import ast
import importlib.util
import importlib.machinery
from pathlib import Path

from pynecore.transformers.import_lifter import ImportLifterTransformer
from pynecore.transformers.import_normalizer import ImportNormalizerTransformer
from pynecore.transformers.module_property import ModulePropertyTransformer
from pynecore.transformers.persistent_series import PersistentSeriesTransformer
from pynecore.transformers.lib_series import LibrarySeriesTransformer
from pynecore.transformers.series import SeriesTransformer
from pynecore.transformers.persistent import PersistentTransformer
from pynecore.transformers.function_isolation import FunctionIsolationTransformer
from pynecore.transformers.input_transformer import InputTransformer
from pynecore.transformers.safe_convert_transformer import SafeConvertTransformer


class PyneLoader(importlib.machinery.SourceFileLoader):
    """Loader that handles AST transformation"""

    # noinspection PyMethodOverriding
    def source_to_code(self, data, path, *, _optimize=-1):
        """Transform source to code if needed"""
        tree = ast.parse(data)
        path = Path(path)

        # Store file path in AST for transformers
        tree._module_file_path = str(path.resolve())

        # Only transform if it has @pyne decorator
        if (tree.body and isinstance(tree.body[0], ast.Expr) and
                isinstance(cast(ast.Expr, tree.body[0]).value, ast.Constant) and
                isinstance(cast(ast.Constant, cast(ast.Expr, tree.body[0]).value).value, str) and
                '@pyne' in cast(ast.Constant, cast(ast.Expr, tree.body[0]).value).value):

            # Remove test cases from the output, because they can coorupt the output
            transformed = tree
            transformed.body = [node for node in transformed.body
                                if not (isinstance(node, ast.FunctionDef)
                                        and node.name.startswith('__test_') and node.name.endswith('__'))]

            # Transform AST
            transformed = ImportLifterTransformer().visit(transformed)
            transformed = ImportNormalizerTransformer().visit(transformed)
            transformed = PersistentSeriesTransformer().visit(transformed)
            transformed = LibrarySeriesTransformer().visit(transformed)
            transformed = FunctionIsolationTransformer().visit(transformed)
            transformed = ModulePropertyTransformer().visit(transformed)
            transformed = SeriesTransformer().visit(transformed)
            transformed = PersistentTransformer().visit(transformed)
            transformed = InputTransformer().visit(transformed)
            transformed = SafeConvertTransformer().visit(transformed)

            ast.fix_missing_locations(transformed)

            # Debug output if requested
            if os.environ.get('PYNE_AST_DEBUG'):
                print("-" * 100)
                print(f"Transformed {path}:")
                try:
                    from rich.syntax import Syntax  # type: ignore
                    from rich import print as rprint  # type: ignore
                    rprint(Syntax(ast.unparse(transformed), "python", word_wrap=True, line_numbers=False))
                except ImportError:
                    print(ast.unparse(transformed))
                print("-" * 100)
            elif os.environ.get('PYNE_AST_DEBUG_RAW'):
                print(ast.unparse(transformed))

            if os.environ.get('PYNE_AST_SAVE'):
                Path("/tmp/pyne").mkdir(parents=True, exist_ok=True)

                with open(f"/tmp/pyne/{path.stem}.py", "w") as f:
                    f.write(ast.unparse(transformed))

            tree = transformed

        # Let Python handle bytecode caching
        return compile(tree, path, 'exec', optimize=_optimize)


class PyneImportHook:
    """Import hook that uses PyneLoader"""

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def find_spec(self, fullname: str, path, target=None):
        """Find and create module spec"""
        if path is None:
            path = sys.path

        if "." in fullname:
            *_, name = fullname.split(".")
        else:
            name = fullname

        for entry in path:
            if entry == "":
                entry = "."
            py_path = Path(entry) / f"{name}.py"
            if py_path.exists():
                return importlib.util.spec_from_file_location(
                    fullname,
                    py_path,
                    loader=PyneLoader(fullname, str(py_path))
                )
        return None


# Install the import hook
sys.meta_path.insert(0, PyneImportHook())



from typing import cast
import ast


class SeriesTransformer(ast.NodeTransformer):
    """Transform Series type variables in AST"""

    def __init__(self):
        # Mapping of scopes to variables
        self.series_vars: dict[str, dict[str, str]] = {}  # scope -> var_name -> series_name

        # Global SeriesImpl assignments to be added to module level
        self.series_assigns: list[ast.AST] = []

        # Current function tracking
        self.current_function: str | None = None
        self.parent_functions: list[str] = []

        # Tracking created Series
        self.collected_series: set[str] = set()

        # Import tracking
        self.has_series_import: bool = False

    def _register_series(self, var_name: str, scope: str | None = None) -> str:
        """
        Register a Series variable and return its global instance name.

        Args:
            var_name: The variable name to register
            scope: The scope where the variable is defined (default: current scope)

        Returns:
            str: The generated global instance name
        """
        if scope is None:
            scope = self._get_current_scope()

        series_name = f'__series_{scope}_{var_name}__'

        if scope not in self.series_vars:
            self.series_vars[scope] = {}
        self.series_vars[scope][var_name] = series_name
        self.collected_series.add(series_name)

        return series_name

    def _get_current_scope(self) -> str:
        """
        Get current scope path.

        Returns:
            str: The current scope path
        """
        if not self.current_function:
            return ""
        return "_".join(self.parent_functions + [self.current_function])

    def _get_series_in_current_scope(self, var_name: str) -> str | None:
        """
        Get series name from current or parent scopes.

        Args:
            var_name: The variable name to look up

        Returns:
            str or None: The series name if found, None otherwise
        """
        current_scope = self._get_current_scope()
        if current_scope in self.series_vars:
            return self.series_vars[current_scope].get(var_name)
        return None

    # noinspection PyShadowingBuiltins
    def visit_Module(self, node: ast.Module) -> ast.Module:
        """
        Create global SeriesImpl instances and register function-variable mapping.

        Args:
            node: The AST module node

        Returns:
            ast.Module: The transformed module
        """
        node = cast(ast.Module, self.generic_visit(node))

        if not self.collected_series:
            return node

        # Create function-to-variables mapping dictionary
        function_vars_dict = {}

        # First collect variables for all functions
        for scope, vars_dict in self.series_vars.items():
            # Convert underscores to dots in key name only
            function_name = scope.replace('_', '.') if scope else "main"
            # Only add non-empty scopes to the registry
            if scope and vars_dict:
                if function_name not in function_vars_dict:
                    function_vars_dict[function_name] = []

                # Add all series variables from this scope
                for _, global_name in vars_dict.items():
                    function_vars_dict[function_name].append(global_name)

        # Special handling for 'main' function
        main_vars = self.series_vars.get('main', {})
        if main_vars:
            function_vars_dict['main'] = [series_name for _, series_name in main_vars.items()]

        # Create the function variable registry
        function_vars_assign = ast.Assign(
            targets=[ast.Name(id='__series_function_vars__', ctx=ast.Store())],
            value=ast.Dict(
                keys=[ast.Constant(value=k) for k in function_vars_dict],
                values=[
                    ast.List(
                        elts=[ast.Constant(value=var) for var in vars],
                        ctx=ast.Load()
                    )
                    for vars in function_vars_dict.values()
                ]
            )
        )

        # Create SeriesImpl import and instances
        imports = [
            ast.ImportFrom(
                module='pynecore.core.series',
                names=[ast.alias(name='SeriesImpl', asname=None)],
                level=0
            )
        ]

        assignments = [
            ast.Assign(
                targets=[ast.Name(id=name, ctx=ast.Store())],
                value=ast.Call(
                    func=ast.Name(id='SeriesImpl', ctx=ast.Load()),
                    args=[],
                    keywords=[]
                )
            )
            for name in sorted(self.collected_series)
        ]

        # Add the function-variable registry
        assignments.append(function_vars_assign)

        # Insert after docstring if exists
        if (node.body and isinstance(node.body[0], ast.Expr) and
                isinstance(cast(ast.Expr, node.body[0]).value, ast.Constant)):
            node.body = [node.body[0]] + imports + node.body[1:]
        else:
            node.body = imports + node.body

        # Find position of first function or assignment
        insert_index = 0
        for i, stmt in enumerate(node.body):
            if isinstance(stmt, (ast.FunctionDef, ast.Assign, ast.AnnAssign)):
                insert_index = i
                break
            # Skip only docstring and imports
            if not (isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant) and
                    isinstance(stmt.value.value, str) or
                    isinstance(stmt, (ast.Import, ast.ImportFrom))):
                insert_index = i
                break

        # Split body and insert assignments
        pre_body = node.body[:insert_index]
        post_body = node.body[insert_index:]

        node.body = pre_body + assignments + post_body
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        """
        Handle function scope and Series parameters.

        Args:
            node: The function definition node

        Returns:
            ast.FunctionDef: The transformed function
        """
        # Store old state
        old_function = self.current_function

        # Special handling for main function
        if node.name == "main":
            self.current_function = "main"
        else:
            if self.current_function:
                self.parent_functions.append(self.current_function)
            self.current_function = node.name

        # Create parameter initializations
        series_initializations = []
        for arg in node.args.args:
            if arg.annotation and self._is_series_type(arg.annotation):
                series_name = self._register_series(arg.arg)
                # Extract inner type from Series[T]
                if isinstance(arg.annotation, ast.Subscript):
                    arg.annotation = arg.annotation.slice
                else:
                    # If no type parameter, remove annotation
                    arg.annotation = None
                # Create initialization statement
                series_initializations.append(
                    ast.Assign(
                        targets=[ast.Name(id=arg.arg, ctx=ast.Store())],
                        value=ast.Call(
                            func=ast.Attribute(
                                value=ast.Name(id=series_name, ctx=ast.Load()),
                                attr='add',
                                ctx=ast.Load()
                            ),
                            args=[ast.Name(id=arg.arg, ctx=ast.Load())],
                            keywords=[]
                        )
                    )
                )

        # Process function body
        node = cast(ast.FunctionDef, self.generic_visit(node))

        # Find the right position to insert initializations after docstring if exists
        insert_pos = 0
        if (node.body and isinstance(node.body[0], ast.Expr) and
                isinstance(cast(ast.Expr, node.body[0]).value, ast.Constant) and
                isinstance(cast(ast.Constant, cast(ast.Expr, node.body[0]).value).value, str)):
            insert_pos = 1

        # Insert series initializations after docstring
        if series_initializations:
            node.body[insert_pos:insert_pos] = series_initializations

        # Restore function context
        if self.parent_functions:
            self.current_function = self.parent_functions.pop()
        else:
            self.current_function = old_function

        return node

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.AST | ast.Assign | None:
        """
        Handle Series type annotations and first value assignment.

        Args:
            node: The annotated assignment node

        Returns:
            AST node: The transformed node
        """
        if not isinstance(node.target, ast.Name):
            return node

        if self._is_series_type(node.annotation):
            var_name = node.target.id
            series_name = self._register_series(var_name)

            if node.value is None:
                return None

            # First assignment uses add()
            return ast.Assign(
                targets=[ast.Name(id=var_name, ctx=ast.Store())],
                value=ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(id=series_name, ctx=ast.Load()),
                        attr='add',
                        ctx=ast.Load()
                    ),
                    args=[self.visit(cast(ast.AST, node.value))],
                    keywords=[]
                )
            )

        return node

    def visit_Assign(self, node: ast.Assign) -> ast.AST | ast.Assign:
        """
        Handle Series value assignments using set().

        Args:
            node: The assignment node

        Returns:
            AST node: The transformed node
        """
        if len(node.targets) != 1 or not isinstance(node.targets[0], ast.Name):
            return self.generic_visit(node)

        var_name = cast(ast.Name, node.targets[0]).id
        series_name = self._get_series_in_current_scope(var_name)

        if series_name:
            # Regular assignment uses set()
            return ast.Assign(
                targets=[ast.Name(id=var_name, ctx=ast.Store())],
                value=ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(id=series_name, ctx=ast.Load()),
                        attr='set',
                        ctx=ast.Load()
                    ),
                    args=[self.visit(cast(ast.AST, node.value))],
                    keywords=[]
                )
            )

        return self.generic_visit(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> ast.AST | ast.Assign:
        """
        Handle augmented assignments (+=, -=, etc).

        Args:
            node: The augmented assignment node

        Returns:
            AST node: The transformed node
        """
        if not isinstance(node.target, ast.Name):
            return self.generic_visit(node)

        var_name = node.target.id
        series_name = self._get_series_in_current_scope(var_name)

        if series_name:
            # Convert augmented assignment to set() with operation
            return ast.Assign(
                targets=[ast.Name(id=var_name, ctx=ast.Store())],
                value=ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(id=series_name, ctx=ast.Load()),
                        attr='set',
                        ctx=ast.Load()
                    ),
                    args=[
                        ast.BinOp(
                            left=ast.Name(id=var_name, ctx=ast.Load()),
                            op=node.op,
                            right=self.visit(cast(ast.AST, node.value))
                        )
                    ],
                    keywords=[]
                )
            )

        return self.generic_visit(node)

    def visit_Name(self, node: ast.Name) -> ast.AST | ast.Name:
        """
        Transform Series references - ONLY when parent is indexing.

        Args:
            node: The name node

        Returns:
            AST node: The transformed node
        """
        if isinstance(node.ctx, ast.Load):
            series_name = self._get_series_in_current_scope(node.id)
            if series_name:
                parent = getattr(node, 'parent', None)
                if isinstance(parent, ast.Subscript) and parent.value == node:
                    return ast.Name(id=series_name, ctx=node.ctx)
        return node

    def visit_Subscript(self, node: ast.Subscript) -> ast.AST:
        """
        Set parent for indexing operations.

        Args:
            node: The subscript node

        Returns:
            AST node: The transformed node
        """
        if isinstance(node.value, ast.AST):
            node.value.parent = node
        node.value = self.visit(cast(ast.AST, node.value))
        node.slice = self.visit(cast(ast.AST, node.slice))
        return node

    @staticmethod
    def _is_series_type(annotation: ast.expr) -> bool:
        """
        Check if type annotation is Series.

        Args:
            annotation: The type annotation expression

        Returns:
            bool: True if the type is Series, False otherwise
        """
        if isinstance(annotation, ast.Subscript):
            return (isinstance(annotation.value, ast.Name) and
                    annotation.value.id == 'Series')
        elif isinstance(annotation, ast.Name):
            return annotation.id == 'Series'
        return False

    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.AST | None:
        """
        Handle imports, remove Series import.

        Args:
            node: The import statement

        Returns:
            AST node or None: The transformed node or None if removed
        """
        if node.module and node.module.startswith('pynecore'):
            # Filter out Series from names
            new_names = [name for name in node.names if name.name != 'Series']
            if not new_names:
                # If no names left, remove the entire import
                return None
            # Create new import with remaining names
            node.names = new_names
            return node
        return node



from ..app import app, app_state

__all__ = []


# noinspection PyShadowingBuiltins
@app.command()
def compile(

):
    """
    Compile Pine Script to Python through pynesys.com
    """
    # TODO: Implement the compile command when API is ready



from typing import TypeVar, TypeAlias
from pynecore.types.na import NA
from .series import Series

T = TypeVar('T')

# Public type alias that allows both T and Persistent[T]
Persistent: TypeAlias = T | NA[T] | Series[T]



import sys
from pathlib import Path

import typer

from ..app import app, app_state
from ..utils.error_hook import setup_global_error_logging

from ...providers import available_providers

# Import commands
from . import run, data, compile

__all__ = ['run', 'data', 'compile']


@app.callback()
def setup(
        ctx: typer.Context,
        workdir: Path = typer.Option(
            app_state.workdir,
            "--workdir", "-w",
            envvar="PYNE_WORK_DIR",
            help="Working directory",
            file_okay=False, dir_okay=True,
            resolve_path=True,
        ),
):
    """
    Pyne Command Line Interface
    """
    if ctx.resilient_parsing or ctx.invoked_subcommand is None:
        return
    if any(arg in ('-h', '--help') for arg in sys.argv[1:]):
        return

    typer.echo("")

    # Check if workdir is available
    workdir_existed = Path(workdir).exists()
    if not workdir_existed:
        typer.echo(f"Working directory '{workdir}' does not exist.")
        typer.confirm("Do you want to create it?", abort=True)

        # Create workdir
        Path(workdir).mkdir(parents=True, exist_ok=False)

    # Create scripts directory
    scripts_dir = Path(workdir) / 'scripts' / 'lib'
    scripts_dir.mkdir(parents=True, exist_ok=True)

    # Create demo.py file only if we created the workdir in this run
    if not workdir_existed:
        demo_file = Path(workdir) / 'scripts' / 'demo.py'
        if not demo_file.exists():
            with demo_file.open('w') as f:
                f.write('''"""
@pyne
Simple Pyne Script Demo

A basic demo showing a 12 and 26 period EMA crossover system.
"""
from pynecore import Series
from pynecore.lib import script, input, plot, color, ta


@script.indicator(
    title="Simple EMA Crossover Demo",
    shorttitle="EMA Demo",
    overlay=True
)
def main(
    src: Series[float] = input.source("close", title="Price Source"),
    fast_length: int = input.int(12, title="Fast EMA Length"),
    slow_length: int = input.int(26, title="Slow EMA Length")
):
    """
    A simple EMA crossover demo
    """
    # Calculate EMAs
    fast_ema = ta.ema(src, fast_length)
    slow_ema = ta.ema(src, slow_length)

    # Plot our indicators
    plot(fast_ema, title="Fast EMA", color=color.blue)
    plot(slow_ema, title="Slow EMA", color=color.red)
''')

    # Create data directory
    data_dir = Path(workdir) / 'data'
    data_dir.mkdir(exist_ok=True)
    # Create output and logs directory
    output_dir = Path(workdir) / 'output' / 'logs'
    output_dir.mkdir(parents=True, exist_ok=True)

    # Create config directory
    config_dir = Path(workdir) / 'config'
    config_dir.mkdir(exist_ok=True)

    # Create providers.toml file for all supported providers (if not exists)
    providers_file = config_dir / 'providers.toml'
    if not providers_file.exists():
        with providers_file.open('w') as f:
            for provider in available_providers:
                f.write(f"[{provider}]\n")
                provider_module = __import__(f"pynecore.providers.{provider}", fromlist=[''])
                provider_class = getattr(
                    provider_module,
                    [p for p in dir(provider_module) if p.endswith('Provider')][0]
                )
                for key, value in provider_class.config_keys.items():
                    if key.startswith('#'):  # Comments
                        f.write(f'{key}\n')
                    else:
                        if isinstance(value, str):
                            f.write(f'{key} = "{value}"\n')
                        elif isinstance(value, bool):
                            f.write(f'{key} = {str(value).lower()}\n')
                        elif isinstance(value, int) or isinstance(value, float):
                            f.write(f'{key} = {value}\n')
                        else:
                            raise ValueError(f"Unsupported type for {key}: {type(value)}")
                f.write("\n")

    # Set workdir in app_state
    app_state.workdir = workdir

    # Setup global error logging
    setup_global_error_logging(workdir / "output" / "logs" / "error.log")



import ast
from typing import Dict, Set, cast


class LibrarySeriesTransformer(ast.NodeTransformer):
    """
    AST transformer that prepares library Series variables for the SeriesTransformer.
    When a library variable is used with indexing, it creates a local Series variable
    declaration that SeriesTransformer can then process.
    """

    def __init__(self):
        self.lib_series_vars: Dict[
            str, Dict[str, tuple[str, ast.AST]]] = {}  # module -> {attr -> (local_name, type_annotation)}
        self.used_series: Set[tuple[str, str]] = set()  # (module, attr) pairs that are used as Series
        self.declarations_to_insert: Dict[str, list[ast.AnnAssign]] = {}  # function_name -> declarations
        self.current_function: str | None = None

    @staticmethod
    def _make_series_name(attr: str) -> str:
        """Generate unique name for the local Series variable"""
        # Replace dots with underscores in attr
        attr = attr.replace('.', '_')
        return f'_lib_{attr}'

    @staticmethod
    def _create_attribute_chain(chain: list[str]) -> ast.expr:
        """Create an attribute chain from a list of names"""
        result: ast.expr = ast.Name(id=chain[0], ctx=ast.Load())
        for name in chain[1:]:
            result = ast.Attribute(
                value=result,
                attr=name,
                ctx=ast.Load()
            )
        return result

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        """Track current function and insert Series declarations"""
        self.current_function = node.name

        # Process function body first
        node = cast(ast.FunctionDef, self.generic_visit(node))

        # Insert declarations for used Series at the start of the function
        if self.current_function in self.declarations_to_insert:
            node.body = self.declarations_to_insert[self.current_function] + node.body

        self.current_function = None
        return node

    def process_series_usage(self, module: str, attr_chain: list[str], type_annotation: ast.AST | None = None) -> str:
        """
        Process a Series usage from a library, creating declaration if needed.
        Returns the local variable name to use.
        """
        if module not in self.lib_series_vars:
            self.lib_series_vars[module] = {}

        # Create underscore version for the variable name
        attr_key = '_'.join(attr_chain)
        if attr_key not in self.lib_series_vars[module]:
            local_name = self._make_series_name(attr_key)
            self.lib_series_vars[module][attr_key] = (local_name, type_annotation)

        local_name = self.lib_series_vars[module][attr_key][0]

        # If this Series hasn't been used in current function yet
        if (module, attr_key) not in self.used_series and self.current_function:
            self.used_series.add((module, attr_key))

            # Create Series declaration with proper attribute chain
            decl = ast.AnnAssign(
                target=ast.Name(id=local_name, ctx=ast.Store()),
                annotation=type_annotation or ast.Name(id='Series', ctx=ast.Load()),
                value=self._create_attribute_chain([module] + attr_chain),  # ez már expr típust ad vissza
                simple=1
            )

            # Store declaration to be inserted
            if self.current_function not in self.declarations_to_insert:
                self.declarations_to_insert[self.current_function] = []
            self.declarations_to_insert[self.current_function].append(decl)

        return local_name

    def visit_Subscript(self, node: ast.Subscript) -> ast.AST:
        """Convert library Series access when used with indexing"""

        # Get the full attribute chain
        def get_attribute_chain(_node):
            if isinstance(_node, ast.Name):
                return [_node.id]
            elif isinstance(_node, ast.Attribute):
                return get_attribute_chain(_node.value) + [_node.attr]
            return []

        if isinstance(node.value, ast.Attribute):
            attr_chain = get_attribute_chain(node.value)
            if attr_chain and attr_chain[0] == 'lib':
                # Use the complete chain after 'lib'
                local_name = self.process_series_usage('lib', attr_chain[1:])
                # Replace lib.xxx.yyy[idx] with local_name[idx]
                node.value = ast.Name(id=local_name, ctx=ast.Load())

        return self.generic_visit(node)



from typing import TypeVar, Any, cast

import builtins

import math
import statistics

from typing_extensions import SupportsIndex

from ..utils.export import export
from ..utils.sequence_view import SequenceView

from ..types.na import NA
from ..types.color import Color
from . import order as _order

T = TypeVar('T')
Number = TypeVar('Number', int, float)


# noinspection PyShadowingBuiltins
@export
def abs(id: list[int | float]) -> list[int | float]:
    """
    Returns an array containing the absolute value of each element in the original array.

    :param id: Input array
    :return: Array containing the absolute value of each element in the original array
    """
    return [builtins.abs(v) for v in id]


# noinspection PyShadowingBuiltins
@export
def avg(id: list[Number]) -> float:
    """
    Returns the average value of the elements in the array.

    :param id: Input array
    :return: Average value of the elements in the array
    """
    return sum(id) / len(id)


# noinspection PyShadowingBuiltins
@export
def binary_search(id: list[Any], val: Any) -> int:
    """
    Returns the index of the specified value in the sorted array using binary search.
    If the value is not found, returns -1.
    The array to search must be sorted in ascending order.

    :param id: Input array
    :param val: Value to search for
    :return: Index of the specified value in the sorted array, or -1 if not found
    """
    low = 0
    high = len(id) - 1
    while low <= high:
        mid = (low + high) // 2
        if id[mid] == val:
            return mid
        else:
            if val < id[mid]:
                high = mid - 1
            else:
                low = mid + 1
    return -1


# noinspection PyShadowingBuiltins
@export
def binary_search_leftmost(id: list[Any], val: Any) -> int:
    """
    Returns the index of the specified value in the sorted array using binary search.
    If the value is not found, returns the index of the leftmost element greater than the value.
    The array to search must be sorted in ascending order.

    :param id: Input array
    :param val: Value to search for
    :return: Index of the specified value in the sorted array, or the index of the leftmost element
             greater than the value
    """
    low = 0
    high = len(id) - 1
    while low <= high:
        mid = (low + high) // 2
        if id[mid] == val:
            return mid
        else:
            if val < id[mid]:
                high = mid - 1
            else:
                low = mid + 1
    return low - 1


# noinspection PyShadowingBuiltins
@export
def binary_search_rightmost(id: list[Any], val: Any) -> int:
    """
    Returns the index of the specified value in the sorted array using binary search.
    If the value is not found, returns the index of the rightmost element less than the value.
    The array to search must be sorted in ascending order.

    :param id: Input array
    :param val: Value to search for
    :return: Index of the specified value in the sorted array, or the index of the rightmost element less than the value
    """
    low = 0
    high = len(id) - 1
    while low <= high:
        mid = (low + high) // 2
        if id[mid] == val:
            return mid
        else:
            if val < id[mid]:
                high = mid - 1
            else:
                low = mid + 1
    return high + 1


# noinspection PyShadowingBuiltins
@export
def clear(id: list[Any]) -> None:
    """
    Removes all elements from the array.

    :param id: Input array
    """
    id.clear()


# noinspection PyShadowingBuiltins
@export
def concat(id1: list[T], id2: list[T]) -> list[T]:
    """
    Concatenates two arrays into a single array.

    :param id1: First array
    :param id2: Second array
    :return: Array containing the elements of both input arrays
    """
    id1.extend(id2)
    return id1


# noinspection PyShadowingBuiltins
@export
def copy(id: list[T]) -> list[T]:
    """
    Returns a shallow copy of the array.

    :param id: Input array
    :return: Shallow copy of the array
    """
    return list(id)


# noinspection PyShadowingBuiltins
@export
def covariance(id1: list[Number], id2: list[Number], biased: bool = True) -> float:
    """
    Returns the covariance between the elements in the two arrays.

    :param id1: First input array
    :param id2: Second input array
    :param biased: If True, calculates the biased covariance. If False, calculates the unbiased covariance.
    :return: Covariance between the elements in the two arrays
    """
    assert len(id1) == len(id2), "Input arrays must have the same length!"
    mean1 = statistics.mean(id1)
    mean2 = statistics.mean(id2)
    length = len(id1)
    summ = 0.0
    for v1, v2 in zip(id1, id2):
        summ += (v1 - mean1) * (v2 - mean2)
    return summ / ((length - 1) if not biased else length)


# noinspection PyShadowingBuiltins
@export
def every(id: list[Any]) -> bool:
    """
    Returns true if all elements of the id array are true, false otherwise.

    :param id: Input array
    :return: True if all elements of the id array are true, false otherwise
    """
    return all(id)


# noinspection PyShadowingBuiltins
@export
def fill(id: list[T], value: T, index_from: int = 0, index_to: int | NA = NA(int)) -> None:
    """
    Fills the elements in the array with the specified value.

    :param id: Input array
    :param value: Value to fill
    :param index_from: Index to start filling from
    :param index_to: Index to stop filling at
    """
    if isinstance(index_to, NA):
        index_to = len(id)
    id[index_from:index_to] = [value] * (index_to - index_from)


# noinspection PyShadowingBuiltins
@export
def first(id: list[T]) -> T:
    """
    Returns the first element in the array.

    :param id: Input array
    :return: First element in the array
    """
    if len(id) == 0:
        raise RuntimeError("Cannot get first element of an empty array!")
    return id[0]


# noinspection PyShadowingBuiltins
@export
def from_items(*items: T) -> list[T]:
    """
    Returns an array containing the specified elements.
    NOTE: this is `array.from()` in Pine Script, but `from` is a reserved keyword in Python

    :param items: Elements to include in the array
    :return: Array containing the specified elements
    """
    return list(items)


# noinspection PyShadowingBuiltins
@export
def get(id: list[T] | SequenceView[T], index: int) -> T:
    """
    Returns the element at the specified index in the array.

    :param id: Input array
    :param index: Index of the element to return
    :return: Element at the specified index in the array
    """
    return cast(T, id[index])


# noinspection PyShadowingBuiltins
@export
def includes(id: list[T], value: T) -> bool:
    """
    Returns true if the array contains the specified value, false otherwise.

    :param id: Input array
    :param value: Value to search for
    :return: True if the array contains the specified value, false otherwise
    """
    return value in id


# noinspection PyShadowingBuiltins
@export
def indexof(id: list[T], value: T) -> int:
    """
    Returns the index of the first occurrence of the specified value in the array.

    :param id: Input array
    :param value: Value to search for
    :return: Index of the first occurrence of the specified value in the array
    """
    try:
        return id.index(value)
    except ValueError:
        return -1


# noinspection PyShadowingBuiltins
@export
def insert(id: list[T], index: int, value: T) -> None:
    """
    Inserts the specified value at the specified index in the array.

    :param id: Input array
    :param index: Index to insert the value at
    :param value: Value to insert
    """
    id.insert(index, value)


# noinspection PyShadowingBuiltins
@export
def join(id: list, separator: str) -> str:
    """
    Concatenates the elements in the array into a single string, separated by the specified separator.

    :param id: Input array
    :param separator: Separator to use
    :return: String containing the concatenated elements
    """
    sa = [str(i) for i in id]  # Ensure all elements are strings
    return separator.join(sa)


# noinspection PyShadowingBuiltins
@export
def last(id: list[T]) -> T:
    """
    Returns the last element in the array.

    :param id: Input array
    :return: Last element in the array
    """
    if len(id) == 0:
        raise RuntimeError("Cannot get last element of an empty array!")
    return id[-1]


# noinspection PyShadowingBuiltins
@export
def lastindexof(id: list[T], value: T) -> int:
    """
    Returns the index of the last occurrence of the specified value in the array.

    :param id: Input array
    :param value: Value to search for
    :return: Index of the last occurrence of the specified value in the array
    """
    try:
        return len(id) - 1 - id[::-1].index(value)
    except ValueError:
        return -1


# noinspection PyShadowingBuiltins
@export
def max(id: list[Number]) -> Number:
    """
    Returns the maximum value in the array.

    :param id: Input array
    :return: Maximum value in the array
    """
    return builtins.max(id)


# noinspection PyShadowingBuiltins
@export
def median(id: list[Number]) -> float:
    """
    Returns the median value of the elements in the array.

    :param id: Input array
    :return: Median value of the elements in the array
    """
    return statistics.median(id)


# noinspection PyShadowingBuiltins
@export
def min(id: list[Number]) -> float:
    """
    Returns the minimum value in the array.

    :param id: Input array
    :return: Minimum value in the array
    """
    return builtins.min(id)


# noinspection PyShadowingBuiltins
@export
def mode(id: list[T]) -> T:
    """
    Returns the most frequently occurring element in the array.

    :param id: Input array
    :return: Most frequently occurring element in the array
    """
    return statistics.mode(id)


# TODO: implement new_box()
# TODO: implement new_line()
# TODO: implement new_label()
# TODO: implement new_table()
# TODO: implement new_linefill()

# noinspection PyShadowingNames
@export
def new(size: int = 0, initial_value: T | NA[T] = NA()) -> list[T | NA[T]]:
    """
    Creates a new array of the specified size, with each element initialized to the specified value.

    :param size: Size of the new array
    :param initial_value: Initial value to set for each element in the array
    :return: New array of the specified size
    """
    assert size >= 0, "Size must be >=0!"
    return [initial_value] * size


# noinspection PyShadowingNames
@export
def new_bool(size: int = 0, initial_value: bool | NA = NA(bool)) -> list[bool | NA[bool]]:
    """
    Creates a new array of the specified size, with each element initialized to the specified value.

    :param size: Size of the new array
    :param initial_value: Initial value to set for each element in the array
    :return: New array of the specified size
    """
    assert size >= 0, "Size must be >=0!"
    assert isinstance(initial_value, (bool, NA)), "Initial value must be bool!"
    return [initial_value] * size


# noinspection PyShadowingNames
@export
def new_color(size: int = 0, initial_value: Color | NA = NA(Color)) -> list[Color | NA[Color]]:
    """
    Creates a new array of the specified size, with each element initialized to the specified value.

    :param size: Size of the new array
    :param initial_value: Initial value to set for each element in the array
    :return: New array of the specified size
    """
    assert size >= 0, "Size must be >=0!"
    assert isinstance(initial_value, (Color, NA)), "Initial value must be Color!"
    return [initial_value] * size


# noinspection PyShadowingNames
@export
def new_float(size: int = 0, initial_value: float | int | NA = NA(float)) -> list[float | NA[float]]:
    """
    Creates a new array of the specified size, with each element initialized to the specified value.

    :param size: Size of the new array
    :param initial_value: Initial value to set for each element in the array
    :return: New array of the specified size
    """
    assert size >= 0, "Size must be >=0!"
    assert isinstance(initial_value, (float, int, NA)), "Initial value must be float!"
    if isinstance(initial_value, int):
        initial_value = float(initial_value)
    return [initial_value] * size


# noinspection PyShadowingNames
@export
def new_int(size: int = 0, initial_value: int | NA = NA(int)) -> list[int | NA[int]]:
    """
    Creates a new array of the specified size, with each element initialized to the specified value.

    :param size: Size of the new array
    :param initial_value: Initial value to set for each element in the array
    :return: New array of the specified size
    """
    assert size >= 0, "Size must be >=0!"
    assert isinstance(initial_value, (int, NA)), "Initial value must be int!"
    return [initial_value] * size


# noinspection PyShadowingNames
@export
def new_string(size: int = 0, initial_value: str | NA = NA(str)) -> list[str | NA[str]]:
    """
    Creates a new array of the specified size, with each element initialized to the specified value.

    :param size: Size of the new array
    :param initial_value: Initial value to set for each element in the array
    :return: New array of the specified size
    """
    assert size >= 0, "Size must be >=0!"
    assert isinstance(initial_value, (str, NA)), "Initial value must be str!"
    return [initial_value] * size


# noinspection PyShadowingBuiltins,PyShadowingNames
@export
def percentile_linear_interpolation(id: list[float], percentage: float) -> float:
    """
    Calculate the percentile value using linear interpolation, following TradingView's logic.

    :param id: List of numeric values
    :param percentage: Percentile (0-100, not 0-1)
    :return: Interpolated value at the given percentile
    :raises ValueError: If arr is empty or percentage is not in [0, 100]
    """
    if not id:
        raise ValueError("Input array is empty")
    if not (0 <= percentage <= 100):
        raise ValueError("Percentage must be between 0 and 100")

    sorted_arr = sorted(id)
    n = len(sorted_arr)

    # Calculate position in 1-indexed system
    pos = n * percentage / 100.0

    # Special cases for extreme percentiles
    if pos < 1:
        return sorted_arr[0]
    if pos >= n:
        return sorted_arr[-1]

    # If pos is an integer, average the two adjacent values
    if pos.is_integer():
        pos_int = int(pos)
        return (sorted_arr[pos_int - 1] + sorted_arr[pos_int]) / 2.0

    # For non-integer positions, perform linear interpolation
    lower_index = int(pos)  # floor of pos
    frac = pos - lower_index
    return sorted_arr[lower_index] + frac * (sorted_arr[lower_index + 1] - sorted_arr[lower_index])


# noinspection PyShadowingBuiltins,PyShadowingNames
@export
def percentile_nearest_rank(id: list[float], percentage: float) -> float:
    """
    Calculate the nearest rank percentile without interpolation.

    :param id: List of numeric values
    :param percentage: Percentile (0-100)
    :return: The value at the nearest rank for the specified percentile
    :raises ValueError: If arr is empty or percentage is not between 0 and 100
    """
    if not id:
        raise ValueError("Input array is empty")
    if not (0 <= percentage <= 100):
        raise ValueError("Percentage must be between 0 and 100")

    sorted_arr = sorted(id)
    n = len(sorted_arr)
    if percentage == 0:
        return sorted_arr[0]

    # Calculate the rank using the ceiling function as per the nearest rank method
    rank = math.ceil(percentage * n / 100)
    # Clamp rank to be within the valid range [1, n]
    rank = builtins.max(1, builtins.min(rank, n))
    # Adjust for 0-indexed array: return the (rank-1)th element
    return sorted_arr[rank - 1]


# noinspection PyShadowingBuiltins,PyShadowingNames
@export
def percentrank(id: list[Number], index: int) -> float:
    """
    Returns the percentile rank of the element at the specified index.
    The percentile rank is the percentage of values less than or equal to the value at index.

    :param id: Input array
    :param index: Index of the element to calculate rank for
    :return: Percentile rank (0-100)
    :raises ValueError: If input array is empty or index is out of range
    """
    if not id:
        raise ValueError("Input array is empty")

    if not 0 <= index < len(id):
        raise ValueError("Index out of range")

    # Get value at index
    value = id[index]

    # Count elements less than or equal
    count = builtins.sum(1 for x in id if x <= value)

    # Calculate percentage
    return (count - 1) * 100 / (len(id) - 1)


# noinspection PyShadowingBuiltins
@export
def pop(id: list[T]) -> T:
    """
    Removes the last element from the array and returns it.

    :param id: Input array
    :return: Last element from the array
    """
    return id.pop()


# noinspection PyShadowingBuiltins
@export
def push(id: list[T], value: T) -> None:
    """
    Appends the specified value to the end of the array.

    :param id: Input array
    :param value: Value to append
    """
    id.append(value)


# noinspection PyShadowingBuiltins
@export
def range(id: list[Number]) -> Number:
    """
    Returns the range of the elements in the array.

    :param id: Input array
    :return: Range of the elements in the array
    """
    return max(id) - min(id)


# noinspection PyShadowingBuiltins
@export
def remove(id: list[T], index: int) -> T:
    """
    Removes the element at the specified index from the array.

    :param id: Input array
    :param index: Index of the element to remove
    :return: The removed element
    """
    return id.pop(index)


# noinspection PyShadowingBuiltins
@export
def reverse(id: list[T]) -> None:
    """
    Reverses the order of the elements in the array.

    :param id: Input array
    """
    id.reverse()


# noinspection PyShadowingBuiltins
@export
def set(id: list[T] | SequenceView[T], index: int, value: T) -> None:
    """
    Sets the value of the element at the specified index in the array.

    :param id: Input array
    :param index: Index of the element to set
    :param value: Value to set
    """
    id[index] = value


# noinspection PyShadowingBuiltins
@export
def shift(id: list[T]) -> T:
    """
    Removes the first element from the array and returns it.

    :param id: Input array
    :return: First element from the array
    """
    return id.pop(0)


# noinspection PyShadowingBuiltins
@export
def size(id: list[Any] | SequenceView[Any]) -> int:
    """
    Returns the number of elements in the array.

    :param id: Input array
    :return: Number of elements in the array
    """
    return len(id)


# noinspection PyShadowingBuiltins
@export
def slice(id: list[T], index_from: int, index_to: int) -> SequenceView[T]:
    """
    The function creates a slice from an existing array. If an object from the slice changes, the
    changes are applied to both the new and the original arrays.

    :param id: Input array
    :param index_from: Index to start the sub-array from
    :param index_to: Index to end the sub-array at
    :return: Slice view of the original array
    """
    return SequenceView(id)[index_from:index_to]  # type: ignore


# noinspection PyShadowingBuiltins
@export
def some(id: list[Any]) -> bool:
    """
    Returns true if at least one element of the id array is true, false otherwise.

    :param id: Input array
    :return: True if at least one element of the id array is true, false otherwise
    """
    return any(id)


# noinspection PyShadowingBuiltins
@export
def sort(id: list[int | float | str], order: _order.Order = _order.ascending) -> None:
    """
    Sorts the elements in the array in ascending or descending order.

    :param id: Input array
    :param order: Order to sort the elements in
    """
    id.sort(reverse=order == _order.descending)


# noinspection PyShadowingBuiltins
@export
def sort_indices(id: list[T], order: _order.Order = _order.ascending) -> list[SupportsIndex]:
    """
    Returns an array of indices which, when used to index the original array, will access its elements
    in their sorted order. It does not modify the original array.

    :param id: Input array
    :param order: Order to sort the elements in
    :return: Array of indices to access the elements in their sorted order
    """
    indices = sorted(builtins.range(len(id)), key=id.__getitem__)  # type: ignore
    if order == _order.descending:
        indices.reverse()
    return indices


# noinspection PyShadowingBuiltins,PyShadowingNames
@export
def standardize(id: list[float | int]) -> list[float | int]:
    """
    Standardizes the input array in a Pine Script-like manner:
      1) Uses a left-to-right summation for the mean (population mean).
      2) Uses a second pass for summing squared differences (population variance).
      3) Computes the population standard deviation (divisor = N).
      4) Returns the z-score for each element.
         - If all input elements are integers, it applies thresholding:
             z < -1 -> -1,
             z > 1  -> 1,
             otherwise 0
         - If any element is float, the result is the continuous z-score value.
    This version is bit-by-bit compatible with Pine Script's `standardize()` function.

    :param id: A list of numeric values (int or float).
    :return: A list containing the standardized values.
    """
    n = len(id)
    if n == 0:
        # You can decide how you want to handle the empty list.
        return []

    mean = statistics.mean(id)
    stdev = math.sqrt(statistics.mean([(v - mean) ** 2 for v in id]))
    z_scores = [(v - mean) / stdev for v in id]

    # If all values are integers, apply the thresholding to get -1, 0, or 1.
    if all(isinstance(v, int) for v in id):
        # Pine Script-style integer thresholding
        return [
            -1 if z < -1 else
            1 if z > 1 else
            0
            for z in z_scores
        ]

    # Otherwise, return the continuous z-score values.
    return z_scores


# noinspection PyShadowingBuiltins
@export
def stdev(id: list[Number], biased: bool = True) -> float:
    """
    Returns the standard deviation of the elements in the array.

    :param id: Input array
    :param biased: If True, calculates the biased standard deviation. If False, calculates the
                   unbiased standard deviation.
    :return: Standard deviation of the elements in the array
    """
    if len(id) < 2:
        return 0.0
    if not biased:
        return statistics.stdev(id)
    mean = statistics.mean(id)
    return math.sqrt(statistics.mean([(v - mean) ** 2 for v in id]))


# noinspection PyShadowingBuiltins
@export
def sum(id: list[float | int]) -> float | int:
    """
    Returns the sum of the elements in the array.

    :param id: Input array
    :return: Sum of the elements in the array
    """
    return builtins.sum(id)


# noinspection PyShadowingBuiltins
@export
def unshift(id: list[T], value: T) -> None:
    """
    Prepends the specified value to the beginning of the array.

    :param id: Input array
    :param value: Value to prepend
    """
    id.insert(0, value)


# noinspection PyShadowingBuiltins
@export
def variance(id: list[Number], biased: bool = True) -> float:
    """
    Returns the variance of the elements in the array.

    :param id: Input array
    :param biased: If True, calculates the biased variance. If False, calculates the unbiased variance.
    :return: Variance of the elements in the array
    """
    if not biased:
        return statistics.variance(id)

    length = len(id)
    mean = statistics.mean(id)
    summ = 0.0
    for v in id:
        summ += (v - mean) ** 2
    return summ / length



from __future__ import annotations
from typing import Any, TypeVar, Generic, Type

T = TypeVar('T')


class NA(Generic[T]):
    """
    Class representing NA (Not Available) values.
    """
    __slots__ = ('type',)

    _type_cache: dict[Type, NA] = {}

    # noinspection PyShadowingBuiltins
    def __new__(cls, type: Type[T] | T | None = int) -> NA[T]:
        if type is None:
            return super().__new__(cls)
        try:
            # Use the cached instance if it exists
            return cls._type_cache[type]
        except KeyError:
            # Create a new instance and store it in the cache
            na = super().__new__(cls)
            cls._type_cache[type] = na
            return na

    # noinspection PyShadowingBuiltins
    def __init__(self, type: Type[T] | T | None = int):
        """
        Initialize a new NA value with an optional type parameter.
        The default type is int.
        """
        self.type = type

    def __repr__(self) -> str:
        """
        Return a string representation of the NA value.
        """
        if self.type is None:
            return "NA"
        return f"NA[{self.type.__name__}]"  # type: ignore

    def __str__(self) -> str:
        """
        Return a string representation of the NA value.
        """
        return ""

    def __hash__(self) -> int:
        """
        Return a hash value for the NA value.
        """
        return hash(self.type)

    def __int__(self) -> NA[int]:
        # We solve this with an AST Transformer
        raise TypeError("NA cannot be converted to int")

    def __float__(self) -> NA[float]:
        # We solve this with an AST Transformer
        raise TypeError("NA cannot be converted to float")

    def __bool__(self) -> bool:
        return False

    #
    # Arithmetic operations
    #

    def __neg__(self) -> NA[T]:
        return NA(self.type)

    def __add__(self, _: Any) -> NA[T]:
        return NA(self.type)

    def __radd__(self, _: Any) -> NA[T]:
        return NA(self.type)

    def __sub__(self, _: Any) -> NA[T]:
        return NA(self.type)

    def __rsub__(self, _: Any) -> NA[T]:
        return NA(self.type)

    def __mul__(self, _: Any) -> NA[T]:
        return NA(self.type)

    def __rmul__(self, _: Any) -> NA[T]:
        return NA(self.type)

    def __truediv__(self, _: Any) -> NA[T]:
        return NA(self.type)

    def __rtruediv__(self, _: Any) -> NA[T]:
        return NA(self.type)

    def __mod__(self, _: Any) -> NA[T]:
        return NA(self.type)

    def __rmod__(self, _: Any) -> NA[T]:
        return NA(self.type)

    def __abs__(self) -> NA[T]:
        return NA(self.type)

    #
    # All comparisons should be false
    #

    def __eq__(self, _: Any) -> bool:
        return False

    def __gt__(self, _: Any) -> bool:
        return False

    def __lt__(self, _: Any) -> bool:
        return False

    def __le__(self, _: Any) -> bool:
        return False

    def __ge__(self, _: Any) -> bool:
        return False

    #
    # In contexts
    #

    def __getattr__(self, _: str) -> NA[T]:
        return self

    def __getitem__(self, _: Any) -> NA[T]:
        return self

    def __call__(self, *_, **__) -> NA[T]:
        return self



import re
from zoneinfo import ZoneInfo
from datetime import datetime, UTC
from functools import lru_cache
from ..lib import syminfo

# Standard formats for non-ISO dates
STANDARD_FORMATS = [
    "%d %b %Y %H:%M:%S %z",  # "20 Feb 2020 15:30:00 +0200"
]

# Pine Script specific formats (without timezone)
PINE_FORMATS = [
    "%b %d %Y %H:%M:%S",  # "Feb 01 2020 22:10:05"
    "%d %b %Y %H:%M:%S",  # "04 Dec 1995 00:12:00"
    "%b %d %Y",  # "Feb 01 2020"
    "%d %b %Y",  # "04 Dec 1995"
    "%Y-%m-%d"  # "2020-02-20"
]


def normalize_timezone(datestring: str) -> str:
    """
    Normalize timezone format to be compatible with Python's datetime.
    Converts formats like "+00:00" to "+0000"

    :param datestring: Input date string
    :return: Normalized date string
    """
    tz_match = re.search(r'([+-])(\d{2}):(\d{2})(?:\s|$)', datestring)
    if tz_match:
        sign, hours, minutes = tz_match.groups()
        new_tz = f"{sign}{hours}{minutes}"
        return datestring[:tz_match.start()] + new_tz + datestring[tz_match.end():]
    return datestring


@lru_cache(maxsize=128)
def parse_timezone(timezone: str) -> ZoneInfo:
    """
    Parse timezone string into ZoneInfo object. Supports:
    - IANA timezone names (e.g. "America/New_York")
    - UTC±HHMM format (e.g. "UTC-5", "UTC+0530")
    - GMT±HHMM format (e.g. "GMT-5", "GMT+0530")
    - Raw offset (e.g. "+0530", "-05:00")

    :param timezone: Timezone string
    :return: ZoneInfo object
    :raises ValueError: If timezone format is invalid
    """
    if not timezone:
        timezone = syminfo.timezone

    # Try as IANA timezone first
    try:
        return ZoneInfo(timezone)
    except KeyError:
        pass

    # Parse UTC/GMT±HHMM format with optional colon
    match = re.match(r'^(UTC|GMT)?([+-])(\d{1,2})(?::?(\d{2})?)?$', timezone)
    if not match:
        raise ValueError(
            f"Invalid timezone format: {timezone}. "
            "Use IANA name (e.g. 'America/New_York') or UTC/GMT±HHMM format (e.g. 'UTC-5', 'GMT+0530')"
        )

    prefix, sign, hours, minutes = match.groups()
    offset = int(hours)
    if minutes:
        offset = float(f"{offset}.{int(minutes) / 60:.2f}"[:-1])

    # UTC/GMT+X maps to Etc/GMT-X and vice versa
    zone = f"Etc/GMT{'-' if sign == '+' else '+'}{abs(offset)}"
    return ZoneInfo(zone)


def parse_datestring(datestring: str) -> datetime:
    """
    Parse date string using multiple formats.
    Handles ISO 8601 with microseconds and timezone offsets.
    If no time is supplied, "00:00" is used.
    If no timezone is supplied, GMT+0 is used.

    :param datestring: Date string to parse
    :return: Parsed datetime object
    :raises ValueError: If the date format is invalid
    """
    datestring = datestring.strip()
    if not datestring:
        return datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)

    # Try parsing ISO 8601 style dates first (handles both T and space separator)
    iso_match = re.match(
        r'(\d{4}-\d{2}-\d{2}[T\s]\d{2}:\d{2}:\d{2}(?:\.\d+)?)'  # datetime part
        r'([+-]\d{2}:\d{2})?$',  # timezone part
        datestring
    )
    if iso_match:
        dt_part, tz_part = iso_match.groups()
        if tz_part:
            datestring = normalize_timezone(datestring)
            dt_str = datestring.replace(' ', 'T')  # Normalize to T for parsing
            try:
                return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f%z")
            except ValueError:
                return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S%z")
        else:
            try:
                dt = datetime.strptime(dt_part.replace(' ', 'T'), "%Y-%m-%dT%H:%M:%S.%f")
            except ValueError:
                dt = datetime.strptime(dt_part.replace(' ', 'T'), "%Y-%m-%dT%H:%M:%S")
            return dt.replace(tzinfo=UTC)

    # Extract timezone if present at the end for other formats
    tz_match = re.search(r'\s*((?:UTC|GMT)?[+-]\d+(?::?\d{2})?)\s*$', datestring)
    if tz_match:
        tz = parse_timezone(
            f"UTC{tz_match.group(1)}" if not tz_match.group(1).startswith(('UTC', 'GMT')) else tz_match.group(1))
        datestring = datestring[:tz_match.start()].strip()
    else:
        tz = UTC

    # Try standard formats (with timezone)
    if tz_match:
        normalized = normalize_timezone(f"{datestring} {tz_match.group(1)}")
        for fmt in STANDARD_FORMATS:
            try:
                return datetime.strptime(normalized, fmt)
            except ValueError:
                continue

    # Try Pine formats (without timezone)
    for fmt in PINE_FORMATS:
        try:
            dt = datetime.strptime(datestring, fmt)
            return dt.replace(tzinfo=tz)
        except ValueError:
            continue

    raise ValueError(
        f"Invalid date format: {datestring}\n"
        "Supported formats:\n"
        "- ISO Style: '2020-02-20T15:30:00+02:00', '2025-01-01 01:23:45-05:00'\n"
        "- With fraction: '2024-08-01T04:38:47.731215+00:00'\n"
        "- RFC Style: '20 Feb 2020 15:30:00 GMT+0200'\n"
        "- Simple Pine: 'Feb 01 2020 22:10:05', '2020-02-20'"
    )



from typing import cast
import ast


class PersistentTransformer(ast.NodeTransformer):
    """
    Transform Persistent type annotations and assignments to global variables
    """

    def __init__(self):
        self.persistent_vars: dict[str, dict[str, str]] = {}  # scope -> var_name -> global_name
        self.scope_stack: list[str] = []
        self.current_scope: str = ""
        self.module_level_assigns: list[ast.Assign] = []
        self.scope_vars: dict[str, set[str]] = {}  # Track all referenced vars per scope
        self.modified_vars: dict[str, set[str]] = {}  # Track modified vars per scope
        self.initialized_flags: dict[str, set[str]] = {}  # scope -> set(init_flags)
        # Track locally declared variables in each scope
        self.local_vars: dict[str, set[str]] = {}  # scope -> set of local var names
        # Track variables defined as Persistent in each scope
        self.persistent_declarations: dict[str, set[str]] = {}  # scope -> set(persistent_var_names)

        self.all_persistent_vars = {}
        self.all_local_vars = {}
        self.current_verifying_scope = None  # Track scope during verification

    def _get_scope_persistents(self, var_name: str) -> str | None:
        """Get persistent variable name from current scope or parent scopes"""
        # Check if variable is locally declared in current scope but NOT as Persistent
        if (self.current_scope in self.local_vars and
                var_name in self.local_vars[self.current_scope] and
                (self.current_scope not in self.persistent_declarations or
                 var_name not in self.persistent_declarations[self.current_scope])):
            return None

        # Check current scope first
        if self.current_scope in self.persistent_vars:
            if var_name in self.persistent_vars[self.current_scope]:
                global_name = self.persistent_vars[self.current_scope][var_name]
                return global_name

        # Then check parent scopes
        for i in range(len(self.scope_stack) - 1, -1, -1):
            scope = "_".join(self.scope_stack[:i + 1])  # Keep underscore for variable names

            if scope in self.persistent_vars:
                if var_name in self.persistent_vars[scope]:
                    global_name = self.persistent_vars[scope][var_name]
                    return global_name

        return None

    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.ImportFrom | None:
        """Handle imports, only remove Persistent while keeping other imports"""
        if node.module and node.module.startswith('pynecore'):
            # Filter out Persistent from names
            new_names = [name for name in node.names if name.name != 'Persistent']
            if not new_names:
                # If no names left, remove the entire import
                return None
            # Create new import with remaining names
            node.names = new_names
        return node

    def visit_Module(self, node: ast.Module) -> ast.Module:
        """Add module level assignments before first function or assignment"""
        # Process the entire module first
        node = cast(ast.Module, self.generic_visit(node))

        # Save persistent variable mappings for final verification
        # Format: {(scope, var_name): global_name}
        for scope, vars_map in self.persistent_vars.items():
            for var_name, global_name in vars_map.items():
                self.all_persistent_vars[(scope, var_name)] = global_name

        # Save all local variables per scope for lookup during verification
        for scope, local_vars in self.local_vars.items():
            self.all_local_vars[scope] = set(local_vars)

        # Final check pass: verify all calls have their persistent arguments properly transformed
        node = cast(ast.Module, self._verify_all_call_args(node))

        if not self.module_level_assigns:
            return node

        # Create function variables dictionary
        function_vars_dict: dict[str, list[str]] = {}

        # Collect all variables for each scope
        for scope, vars_dict in self.persistent_vars.items():
            # Register every scope with its variables - convert '_' to '.' in key name only
            function_name = scope.replace('_', '.') if scope else "main"
            function_vars: list[str] = []

            for var_name, global_name in vars_dict.items():
                function_vars.append(global_name)

            # Add initialization flags that actually exist
            if scope in self.initialized_flags:
                for init_flag in self.initialized_flags[scope]:
                    function_vars.append(init_flag)

            if function_vars:  # Only add if there are variables
                function_vars_dict[function_name] = function_vars

        # Create the registration dictionary if there are function variables
        if function_vars_dict:
            # noinspection PyShadowingBuiltins
            function_vars_assign = ast.Assign(
                targets=[ast.Name(id='__persistent_function_vars__', ctx=ast.Store())],
                value=ast.Dict(
                    keys=[ast.Constant(value=k) for k in function_vars_dict],
                    values=[
                        ast.List(
                            elts=[ast.Constant(value=var) for var in vars],
                            ctx=ast.Load()
                        )
                        for vars in function_vars_dict.values()
                    ]
                )
            )
            self.module_level_assigns.append(function_vars_assign)

        # Find position of first function or assignment
        insert_index = 0
        for i, stmt in enumerate(node.body):
            if isinstance(stmt, (ast.FunctionDef, ast.Assign, ast.AnnAssign)):
                insert_index = i
                break
            # Skip only docstring and imports
            if not (isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant) and
                    isinstance(stmt.value.value, str) or
                    isinstance(stmt, (ast.Import, ast.ImportFrom))):
                insert_index = i
                break

        # Split body and insert assignments
        pre_body = node.body[:insert_index]
        post_body = node.body[insert_index:]

        return ast.Module(
            body=pre_body + self.module_level_assigns + post_body,
            type_ignores=node.type_ignores
        )

    def _verify_all_call_args(self, node: ast.AST) -> ast.AST:
        """Recursively verify and fix all Call nodes in the AST"""
        # Track current scope when verifying function definitions
        if isinstance(node, ast.FunctionDef):
            old_scope = self.current_verifying_scope
            if not self.current_verifying_scope:
                self.current_verifying_scope = node.name
            else:
                self.current_verifying_scope = f"{self.current_verifying_scope}_{node.name}"

            # Process function and update scope back when done
            result = self._process_verify_node(cast(ast.FunctionDef, node))
            self.current_verifying_scope = old_scope
            return result

        return self._process_verify_node(node)

    def _process_verify_node(self, node: ast.AST) -> ast.AST:
        """Process a single node during verification"""
        if isinstance(node, ast.Call):
            # Process all arguments to ensure they are transformed
            for i, arg in enumerate(node.args):
                if isinstance(arg, ast.Name):
                    var_name = arg.id
                    # Skip variables that are already transformed
                    if var_name.startswith('__persistent_'):
                        continue

                    # Check if this is a local variable in current scope
                    if (self.current_verifying_scope and
                            self.current_verifying_scope in self.all_local_vars and
                            var_name in self.all_local_vars[self.current_verifying_scope]):
                        # It's a local variable, don't transform it
                        continue

                    # Look for this variable in persistent vars, checking current scope first
                    global_name = None
                    if self.current_verifying_scope:
                        scope_key = (self.current_verifying_scope, var_name)
                        if scope_key in self.all_persistent_vars:
                            global_name = self.all_persistent_vars[scope_key]

                    # Then check parent scopes
                    if not global_name and self.current_verifying_scope:
                        # Try parent scopes
                        scope_parts = self.current_verifying_scope.split('_')
                        for _ in range(len(scope_parts) - 1, 0, -1):
                            parent_scope = '_'.join(scope_parts[:i])
                            parent_key = (parent_scope, var_name)
                            if parent_key in self.all_persistent_vars:
                                global_name = self.all_persistent_vars[parent_key]
                                break

                    # If found, transform
                    if global_name:
                        node.args[i] = ast.Name(id=global_name, ctx=ast.Load())

            # Process keyword arguments too
            for kw in node.keywords:
                if isinstance(kw.value, ast.Name):
                    var_name = kw.value.id
                    # Skip variables that are already transformed
                    if var_name.startswith('__persistent_'):
                        continue

                    # Check if this is a local variable in current scope
                    if (self.current_verifying_scope and
                            self.current_verifying_scope in self.all_local_vars and
                            var_name in self.all_local_vars[self.current_verifying_scope]):
                        # It's a local variable, don't transform it
                        continue

                    # Look for this variable in persistent vars, checking current scope first
                    global_name = None
                    if self.current_verifying_scope:
                        scope_key = (self.current_verifying_scope, var_name)
                        if scope_key in self.all_persistent_vars:
                            global_name = self.all_persistent_vars[scope_key]

                    # Then check parent scopes
                    if not global_name and self.current_verifying_scope:
                        # Try parent scopes
                        scope_parts = self.current_verifying_scope.split('_')
                        for i in range(len(scope_parts) - 1, 0, -1):
                            parent_scope = '_'.join(scope_parts[:i])
                            parent_key = (parent_scope, var_name)
                            if parent_key in self.all_persistent_vars:
                                global_name = self.all_persistent_vars[parent_key]
                                break

                    # If found, transform
                    if global_name:
                        kw.value = ast.Name(id=global_name, ctx=ast.Load())

        # Continue with recursion for all child nodes
        for field, old_value in ast.iter_fields(node):
            if isinstance(old_value, list):
                new_values = []
                for value in old_value:
                    if isinstance(value, ast.AST):
                        value = self._verify_all_call_args(value)
                        new_values.append(value)
                    else:
                        new_values.append(value)
                old_value[:] = new_values
            elif isinstance(old_value, ast.AST):
                new_node = self._verify_all_call_args(old_value)
                setattr(node, field, new_node)
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        """Process function definitions"""
        if node.name == "main":
            # Handle main function specially
            self.scope_stack.append("main")
        else:
            self.scope_stack.append(node.name)

        self.current_scope = "_".join(self.scope_stack)  # Keep underscore for variable names

        # Initialize tracking sets for this scope
        self.scope_vars.setdefault(self.current_scope, set())
        self.modified_vars.setdefault(self.current_scope, set())
        self.initialized_flags.setdefault(self.current_scope, set())
        self.local_vars.setdefault(self.current_scope, set())
        self.persistent_declarations.setdefault(self.current_scope, set())

        # Add function arguments to local variables
        for arg in node.args.args:
            self.local_vars[self.current_scope].add(arg.arg)

        # Process function body
        node = cast(ast.FunctionDef, self.generic_visit(node))

        # *** FŐ MÓDOSÍTÁS: Csak azokat a változókat adjuk hozzá a global utasításhoz,
        # amelyeket ténylegesen módosítunk (írunk) a függvényben ***
        globals_to_declare = set()

        # A módosított változók közül csak a perzisztens változókat és inicializálási flageket adjuk hozzá
        if self.current_scope in self.modified_vars:
            for global_name in self.modified_vars[self.current_scope]:
                if global_name.startswith('__persistent_'):
                    globals_to_declare.add(global_name)

                    # Az inicializálási flageket is hozzá kell adnunk, ha kapcsolódnak
                    # módosított perzisztens változóhoz
                    init_flag = f"{global_name}_initialized__"
                    if init_flag in self.initialized_flags.get(self.current_scope, set()):
                        globals_to_declare.add(init_flag)

        # Csak akkor adjuk hozzá a globális deklarációt, ha van mit deklarálni
        if globals_to_declare:
            insert_pos = 0
            if (node.body and isinstance(node.body[0], ast.Expr) and
                    isinstance(cast(ast.Expr, node.body[0]).value, ast.Constant) and
                    isinstance(cast(ast.Constant, cast(ast.Expr, node.body[0]).value).value, str)):
                insert_pos = 1

            node.body.insert(insert_pos, ast.Global(names=sorted(globals_to_declare)))

        # Fix nonlocal statements - remove persistent variables
        persistent_globals = set()
        for i in range(len(self.scope_stack)):
            scope = "_".join(self.scope_stack[:i + 1])
            if scope in self.persistent_vars:
                for var_name, global_name in self.persistent_vars[scope].items():
                    if global_name in self.modified_vars.get(self.current_scope, set()):
                        persistent_globals.add(var_name)

        new_body = []
        for stmt in node.body:
            if isinstance(stmt, ast.Nonlocal):
                # Filter out variables that are now globals
                new_names = [name for name in stmt.names if name not in persistent_globals]
                if new_names:
                    stmt.names = new_names
                    new_body.append(stmt)
                # Skip empty nonlocal statements
            else:
                new_body.append(stmt)
        node.body = new_body

        self.scope_stack.pop()
        self.current_scope = "_".join(self.scope_stack)  # Keep underscore for variable names
        return node

    @staticmethod
    def _is_persistent_type(annotation: ast.expr) -> bool:
        """Check if the annotation is any form of Persistent type"""
        if isinstance(annotation, ast.Name):
            # Simple case: Persistent
            return annotation.id == 'Persistent'
        elif isinstance(annotation, ast.Subscript):
            # Persistent[T] case
            if isinstance(annotation.value, ast.Name):
                return annotation.value.id == 'Persistent'
        elif isinstance(annotation, ast.Attribute):
            # module.Persistent case
            return annotation.attr == 'Persistent'
        return False

    @staticmethod
    def _is_literal_or_na(node: ast.expr) -> bool:
        """Check if a node represents a literal value or na"""
        if isinstance(node, ast.Constant):
            return True
        if isinstance(node, ast.Name):
            return node.id == 'na'
        return False

    def visit_Call(self, node: ast.Call) -> ast.AST:
        """Handle function calls and ensure arguments get proper transformation"""
        # Visit children first (function, args, keywords) to process nested calls
        # IMPORTANT: store the result, or nested transformations might be lost
        visited_node = cast(ast.Call, self.generic_visit(node))

        # Now transform args
        for i, arg in enumerate(visited_node.args):
            if isinstance(arg, ast.Name):
                var_name = arg.id
                global_name = self._get_scope_persistents(var_name)

                if global_name:
                    # Replace argument with global name
                    visited_node.args[i] = ast.Name(id=global_name, ctx=ast.Load())

                    # Track usage
                    if self.current_scope:
                        self.scope_vars.setdefault(self.current_scope, set())
                        self.scope_vars[self.current_scope].add(global_name)

        # Also handle keyword arguments
        for kw in visited_node.keywords:
            if isinstance(kw.value, ast.Name):
                var_name = kw.value.id
                global_name = self._get_scope_persistents(var_name)

                if global_name:
                    # Replace keyword value with global name
                    kw.value = ast.Name(id=global_name, ctx=ast.Load())

                    # Track usage
                    if self.current_scope:
                        self.scope_vars.setdefault(self.current_scope, set())
                        self.scope_vars[self.current_scope].add(global_name)

        return visited_node

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.AST | None:
        """Convert any Persistent type annotated assignments"""
        if not isinstance(node.target, ast.Name):
            return node

        if self._is_persistent_type(node.annotation):
            var_name = node.target.id

            # Mark this variable as a Persistent declaration in this scope
            self.persistent_declarations.setdefault(self.current_scope, set())
            self.persistent_declarations[self.current_scope].add(var_name)

            # Add to local vars to track the variable in this scope
            self.local_vars.setdefault(self.current_scope, set())
            self.local_vars[self.current_scope].add(var_name)

            # Generate global name using current scope
            global_name = f"__persistent_{self.current_scope}_{var_name}__"

            # Initialize scope dict if needed
            if self.current_scope not in self.persistent_vars:
                self.persistent_vars[self.current_scope] = {}

            # Store the mapping
            self.persistent_vars[self.current_scope][var_name] = global_name

            # Track this variable in current scope
            if self.current_scope:
                if self.current_scope not in self.scope_vars:
                    self.scope_vars[self.current_scope] = set()
                self.scope_vars[self.current_scope].add(global_name)

                # *** JAVÍTÁS: Csak akkor jelöljük módosítottként, ha nem literál az érték ***
                if node.value and not self._is_literal_or_na(node.value):
                    if self.current_scope not in self.modified_vars:
                        self.modified_vars[self.current_scope] = set()
                    self.modified_vars[self.current_scope].add(global_name)

            # Handle module level assignments and initialization
            if node.value:
                if self._is_literal_or_na(node.value):
                    # For literals, just add module level assignment
                    self.module_level_assigns.append(
                        ast.Assign(
                            targets=[ast.Name(id=global_name, ctx=ast.Store())],
                            value=node.value
                        )
                    )
                else:
                    # For non-literal values:
                    # 1. Initialize with None at module level
                    self.module_level_assigns.append(
                        ast.Assign(
                            targets=[ast.Name(id=global_name, ctx=ast.Store())],
                            value=ast.Constant(value=None)
                        )
                    )
                    # 2. Add initialization flag
                    init_flag = f"{global_name}_initialized__"
                    self.module_level_assigns.append(
                        ast.Assign(
                            targets=[ast.Name(id=init_flag, ctx=ast.Store())],
                            value=ast.Constant(value=False)
                        )
                    )
                    # 3. Register the initialization flag
                    if self.current_scope:
                        if self.current_scope not in self.scope_vars:
                            self.scope_vars[self.current_scope] = set()
                        self.scope_vars[self.current_scope].add(init_flag)

                        # Mark initialization flag as modified since we'll be writing to it
                        if self.current_scope not in self.modified_vars:
                            self.modified_vars[self.current_scope] = set()
                        self.modified_vars[self.current_scope].add(init_flag)

                        if self.current_scope not in self.initialized_flags:
                            self.initialized_flags[self.current_scope] = set()
                        self.initialized_flags[self.current_scope].add(init_flag)

                    return cast(ast.AST, ast.If(
                        test=ast.UnaryOp(
                            op=ast.Not(),
                            operand=ast.Name(id=init_flag, ctx=ast.Load())
                        ),
                        body=[
                            ast.Assign(
                                targets=[ast.Name(id=global_name, ctx=ast.Store())],
                                value=self.visit(cast(ast.AST, node.value))
                            ),
                            ast.Assign(
                                targets=[ast.Name(id=init_flag, ctx=ast.Store())],
                                value=ast.Constant(value=True)
                            )
                        ],
                        orelse=[]
                    ))
            else:
                # No initial value, initialize with na
                self.module_level_assigns.append(
                    ast.Assign(
                        targets=[ast.Name(id=global_name, ctx=ast.Store())],
                        value=ast.Name(id='na', ctx=ast.Load())
                    )
                )
            return None
        return node

    def visit_Assign(self, node: ast.Assign) -> ast.Assign:
        """Convert normal assignments to persistent variables"""
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            target = cast(ast.Name, node.targets[0])
            var_name = target.id

            # First, check if there are any Persistent declarations in the current scope
            is_first_assignment = (var_name not in self.local_vars.get(self.current_scope, set()))

            # If this is the first assignment in this scope, mark it as a local variable
            if is_first_assignment:
                self.local_vars.setdefault(self.current_scope, set())
                self.local_vars[self.current_scope].add(var_name)

            # Now check if it's a persistent variable reference
            global_name = self._get_scope_persistents(var_name)

            if global_name:
                # Track this variable in current scope
                if self.current_scope:
                    self.scope_vars.setdefault(self.current_scope, set())
                    self.scope_vars[self.current_scope].add(global_name)

                    # Mark variable as modified because we're assigning to it
                    self.modified_vars.setdefault(self.current_scope, set())
                    self.modified_vars[self.current_scope].add(global_name)

                # Visit the value part first to transform any references in it
                transformed_value = self.visit(cast(ast.AST, node.value))

                return ast.Assign(
                    targets=[ast.Name(id=global_name, ctx=ast.Store())],
                    value=transformed_value
                )

        # If not a persistent assignment, still visit the value part
        node.value = self.visit(cast(ast.AST, node.value))
        return node

    def visit_AugAssign(self, node: ast.AugAssign) -> ast.AugAssign:
        """Handle augmented assignments (+=, *=, etc.)"""
        if isinstance(node.target, ast.Name):
            var_name = node.target.id
            global_name = self._get_scope_persistents(var_name)

            if global_name and self.current_scope:
                # Mark as modified since we're augmenting it
                self.modified_vars.setdefault(self.current_scope, set())
                self.modified_vars[self.current_scope].add(global_name)

                # Replace with the global name
                node.target = ast.Name(id=global_name, ctx=ast.Store())
                node.value = self.visit(cast(ast.AST, node.value))
                return node

        return cast(ast.AugAssign, self.generic_visit(node))

    def visit_Name(self, node: ast.Name) -> ast.Name:
        """Convert variable references using scope-aware lookup"""
        var_name = node.id
        global_name = self._get_scope_persistents(var_name)

        if global_name:
            # Track this variable in current scope if it's being used
            if self.current_scope:
                self.scope_vars.setdefault(self.current_scope, set())
                self.scope_vars[self.current_scope].add(global_name)

                # Only mark as modified if it's in a Store context
                if isinstance(node.ctx, ast.Store):
                    self.modified_vars.setdefault(self.current_scope, set())
                    self.modified_vars[self.current_scope].add(global_name)

            return ast.Name(id=global_name, ctx=node.ctx)
        return node



from typing import TypeVar, cast
import builtins
import math

from ..types.na import NA
from ..types.series import Series

from ..utils.export import export

from . import syminfo

from ..core.series import SeriesImpl as _SeriesImpl
from ..core.random import PineRandom as _PineRandom

TFI = TypeVar('TFI', float, int)

# Constants
e = math.e
pi = math.pi
phi = (1 + math.sqrt(5)) / 2
rphi = 1 / phi

export('e', 'pi', 'phi', 'rphi')

__persistent_function_vars__ = {}
__series_function_vars__ = {}


# noinspection PyShadowingBuiltins
@export
def abs(number: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the absolute value of a number.

    :param number: A number.
    :return: The absolute value of the number.
    """
    if isinstance(number, NA):
        return NA(float)
    return builtins.abs(number)


@export
def acos(value: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the arc cosine of a value.

    :param value: A value.
    :return: The arc cosine of the value.
    """
    if isinstance(value, NA):
        return NA(float)
    return math.acos(value)


@export
def asin(value: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the arc sine of a value.

    :param value: A value.
    :return: The arc sine of the value.
    """
    if isinstance(value, NA):
        return NA(float)
    return math.asin(value)


@export
def atan(value: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the arc tangent of a value.

    :param value: A value.
    :return: The arc tangent of the value.
    """
    if isinstance(value, NA):
        return NA(float)
    return math.atan(value)


@export
def avg(*numbers: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the average of the numbers.

    :param numbers: Numbers.
    :return: The average of the numbers.
    """
    assert numbers, "At least one number is necessary!"

    if any(isinstance(n, NA) for n in numbers):
        return NA(float)

    return builtins.sum(cast(TFI, n) for n in numbers) / len(numbers)


@export
def ceil(number: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the smallest integer greater than or equal to a number.

    :param number: A number.
    :return: The smallest integer greater than or equal to the number.
    """
    if isinstance(number, NA):
        return NA(float)
    return math.ceil(number)


@export
def cos(angle: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the cosine of an angle.

    :param angle: An angle in radians.
    :return: The cosine of the angle.
    """
    if isinstance(angle, NA):
        return NA(float)
    return math.cos(angle)


@export
def exp(number: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns e raised to the power of a number.

    :param number: A number.
    :return: e raised to the power of the number.
    """
    if isinstance(number, NA):
        return NA(float)
    return math.exp(number)


@export
def floor(number: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the largest integer less than or equal to a number.

    :param number: A number.
    :return: The largest integer less than or equal to the number.
    """
    if isinstance(number, NA):
        return NA(float)
    return int(number)


@export
def log(number: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the natural logarithm of a number.

    :param number: A number.
    :return: The natural logarithm of the number.
    """
    if isinstance(number, NA):
        return NA(float)
    return math.log(number)


@export
def log10(number: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the base-10 logarithm of a number.

    :param number: A number.
    :return: The base-10 logarithm of the number.
    """
    if isinstance(number, NA):
        return NA(float)
    return math.log10(number)


# noinspection PyShadowingBuiltins
@export
def max(*numbers: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the largest number.

    :param numbers: Numbers.
    :return: The largest number.
    """
    assert numbers, "At least one number is necessary!"

    if any(isinstance(n, NA) for n in numbers):
        return NA(float)

    return builtins.max(cast(list[TFI], numbers))


# noinspection PyShadowingBuiltins
@export
def min(*numbers: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the smallest number.

    :param numbers: Numbers.
    :return: The smallest number.
    """
    assert numbers, "At least one number is necessary!"

    if any(isinstance(n, NA) for n in numbers):
        return NA(float)

    return builtins.min(cast(list[TFI], numbers))


# noinspection PyShadowingBuiltins
@export
def pow(base: TFI | NA[TFI], exponent: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns a number raised to the power of another number.

    :param base: The base number.
    :param exponent: The exponent number.
    :return: The base number raised to the power of the exponent number.
    """
    if isinstance(base, NA) or isinstance(exponent, NA):
        return NA(float)

    return base ** exponent


__persistent_random_prng__: _PineRandom | None = None
__persistent_function_vars__['random'] = ['__persistent_random_prng__']


# noinspection PyShadowingBuiltins,PyShadowingNames
@export
def random(min: TFI | NA[TFI] = 0, max: TFI | NA[TFI] = 1, seed: int | NA[int] = NA(int)) -> float | NA[float]:
    """
    Returns a random number between two numbers.

    :param min: The minimum number.
    :param max: The maximum number.
    :param seed: The seed for the random number generator.
    :return: A random number between the minimum and maximum numbers.
    """
    global __persistent_random_prng__
    if __persistent_random_prng__ is None:
        __persistent_random_prng__ = _PineRandom(seed)
    res = __persistent_random_prng__.random(min, max)
    return res


# noinspection PyShadowingBuiltins
@export
def round(number: TFI | NA[TFI], precision: int | NA[int] = NA(int)) -> float | NA[float]:
    """
    Returns a number rounded to a specified number of decimal places.

    :param number: A number.
    :param precision: The number of decimal places to round to.
    :return: The rounded number.
    """
    if isinstance(number, NA):
        return NA(float)
    print(f"number: {number}, precision: {precision}")
    if isinstance(precision, NA):
        return builtins.round(number)
    return builtins.round(number, precision)


@export
def round_to_mintick(number: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns value rounded to symbol's mintick with ties rounding up.
    """
    return int(number / syminfo.mintick + 0.5) * syminfo.mintick


@export
def sign(number: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the sign of a number.

    :param number: A number.
    :return: The sign of the number.
    """
    if isinstance(number, NA):
        return NA(float)
    if number == 0.0:
        return 0.0
    if number > 0.0:
        return 1.0
    return -1.0


@export
def sin(angle: float | int | NA) -> float | NA[float]:
    """
    Returns the sine of an angle.

    :param angle: An angle in radians.
    :return: The sine of the angle.
    """
    if isinstance(angle, NA):
        return NA(float)
    return math.sin(angle)


@export
def sqrt(number: float | int | NA) -> float | NA[float]:
    """
    Returns the square root of a number.

    :param number: A number.
    :return: The square root of the number.
    """
    if isinstance(number, NA):
        return NA(float)
    try:
        return math.sqrt(number)
    except ValueError:
        return NA(float)


__series_summ_source__: _SeriesImpl[float | NA[float]] = _SeriesImpl()
__persistent_summ_summ__: float = 0.0
__persistent_summ_count__: int = 0
__persistent_function_vars__['sum'] = ['__persistent_summ_summ__', '__persistent_summ_count__']
__series_function_vars__['sum'] = ['__series_summ_source__']


# noinspection PyShadowingBuiltins
@export
def sum(source: Series[TFI | NA[TFI]], length: int) -> float | NA[float] | Series[TFI | NA[TFI]]:
    """
    Returns the sum of a series over a specified length.

    :param source: Source series
    :param length: Length of the sum
    :return: The sliding sum of the series
    """
    global __series_summ_source__, __persistent_summ_summ__, __persistent_summ_count__

    if length == 1:  # Shortcut
        return source
    assert length > 0, "Invalid length, length must be greater than 0!"

    isna = isinstance(source, NA)
    if not isna:
        __series_summ_source__.add(source)

    if __persistent_summ_count__ < length - 1:
        if not isna:
            __persistent_summ_count__ += 1
            __persistent_summ_summ__ += source
        return NA(float)
    elif __persistent_summ_count__ == length - 1:
        if isna:
            return NA(float)
        __persistent_summ_count__ += 1
    else:
        if isna:
            return __persistent_summ_summ__
        __persistent_summ_summ__ -= float(__series_summ_source__[length])

    __persistent_summ_summ__ += source

    return __persistent_summ_summ__


@export
def tan(angle: TFI | NA[TFI]) -> float | NA[float]:
    """
    Returns the tangent of an angle.

    :param angle: An angle in radians.
    :return: The tangent of the angle.
    """
    if isinstance(angle, NA):
        return NA(float)
    return math.tan(angle)


@export
def todegrees(angle: TFI | NA[TFI]) -> float | NA[float]:
    """
    Converts an angle from radians to degrees.

    :param angle: An angle in radians.
    :return: The angle in degrees.
    """
    if isinstance(angle, NA):
        return NA(float)
    return math.degrees(angle)


@export
def toradians(angle: TFI | NA[TFI]) -> float | NA[float]:
    """
    Converts an angle from degrees to radians.

    :param angle: An angle in degrees.
    :return: The angle in radians.
    """
    if isinstance(angle, NA):
        return NA(float)
    return math.radians(angle)
